# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.entities.acis import (
    Body,
    Solid3d,
    Region,
    Surface,
    ExtrudedSurface,
    LoftedSurface,
    RevolvedSurface,
    SweptSurface,
)
from ezdxf.entities import DXFEntity
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.lldxf.tags import Tags
from ezdxf.lldxf.const import DXF2010

BODY = """0
BODY
5
0
330
0
100
AcDbEntity
8
0
100
AcDbModelerGeometry
70
1
"""


class MockDoc:
    def __init__(self):
        self.dxfversion = DXF2010


@pytest.fixture
def entity():
    return Body.from_text(BODY, doc=MockDoc())


def special_export(entity: DXFEntity, dxfversion=DXF2010):
    """These ACIS entities do not contain any ACIS data, which the regular export method
    does not export.
    """
    collector = TagCollector(dxfversion=dxfversion)
    entity.export_base_class(collector)
    entity.export_entity(collector)
    return Tags(collector.tags)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "BODY" in ENTITY_CLASSES
    assert "3DSOLID" in ENTITY_CLASSES
    assert "REGION" in ENTITY_CLASSES
    assert "SURFACE" in ENTITY_CLASSES
    assert "EXTRUDEDSURFACE" in ENTITY_CLASSES
    assert "LOFTEDSURFACE" in ENTITY_CLASSES
    assert "REVOLVEDSURFACE" in ENTITY_CLASSES
    assert "SWEPTSURFACE" in ENTITY_CLASSES


def test_default_init():
    entity = Body()
    assert entity.dxftype() == "BODY"

    entity = Solid3d()
    assert entity.dxftype() == "3DSOLID"

    entity = Region()
    assert entity.dxftype() == "REGION"

    entity = Surface()
    assert entity.dxftype() == "SURFACE"

    entity = ExtrudedSurface()
    assert entity.dxftype() == "EXTRUDEDSURFACE"

    entity = LoftedSurface()
    assert entity.dxftype() == "LOFTEDSURFACE"

    entity = RevolvedSurface()
    assert entity.dxftype() == "REVOLVEDSURFACE"

    entity = SweptSurface()
    assert entity.dxftype() == "SWEPTSURFACE"


def test_default_new():
    entity = Body.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": 7,
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.version == 1


def test_regular_dxf_export_of_empty_acis_entities():
    entity = Body.from_text(BODY, doc=MockDoc())
    assert (
        len(TagCollector.dxftags(entity, DXF2010)) == 0
    ), "do not export ACIS entities without content"


def test_body_write_dxf():
    entity = Body.from_text(BODY, doc=MockDoc())
    result = special_export(entity)
    expected = basic_tags_from_text(BODY)
    assert result == expected


def test_region_write_dxf():
    entity = Region.from_text(REGION, doc=MockDoc())
    result = special_export(entity, dxfversion=DXF2010)
    expected = basic_tags_from_text(REGION)
    assert result == expected


REGION = """0
REGION
5
0
330
0
100
AcDbEntity
8
0
100
AcDbModelerGeometry
70
1
"""


def test_3dsolid_write_dxf():
    entity = Solid3d.from_text(SOLID3D, doc=MockDoc())
    result = special_export(entity, dxfversion=DXF2010)
    expected = basic_tags_from_text(SOLID3D)
    assert result == expected


SOLID3D = """0
3DSOLID
5
0
330
0
100
AcDbEntity
8
0
100
AcDbModelerGeometry
70
1
100
AcDb3dSolid
350
0
"""


def test_surface_write_dxf():
    entity = Surface.from_text(SURFACE, doc=MockDoc())
    result = special_export(entity, dxfversion=DXF2010)
    expected = basic_tags_from_text(SURFACE)
    assert result == expected


SURFACE = """0
SURFACE
5
0
330
0
100
AcDbEntity
8
0
100
AcDbModelerGeometry
70
1
100
AcDbSurface
71
0
72
0
"""


def test_extruded_surface_write_dxf():
    entity = ExtrudedSurface.from_text(EXTRUDEDSURFACE, doc=MockDoc())
    result = special_export(entity, dxfversion=DXF2010)
    expected = basic_tags_from_text(EXTRUDEDSURFACE)
    assert result == expected


EXTRUDEDSURFACE = """0
EXTRUDEDSURFACE
5
0
330
0
100
AcDbEntity
8
0
100
AcDbModelerGeometry
70
1
100
AcDbSurface
71
0
72
0
100
AcDbExtrudedSurface
90
18
10
0.0
20
0.0
30
0.0
40
1.0
40
0.0
40
0.0
40
0.0
40
0.0
40
1.0
40
0.0
40
0.0
40
0.0
40
0.0
40
1.0
40
0.0
40
0.0
40
0.0
40
0.0
40
1.0
42
0.0
43
0.0
44
0.0
45
0.0
48
1.0
49
0.0
46
1.0
46
0.0
46
0.0
46
0.0
46
0.0
46
1.0
46
0.0
46
0.0
46
0.0
46
0.0
46
1.0
46 
0.0
46
0.0
46
0.0
46
0.0
46
1.0
47
1.0
47
0.0
47
0.0
47
0.0
47
0.0
47
1.0
47
0.0
47
0.0
47
0.0
47
0.0
47
1.0
47
0.0
47
0.0
47
0.0
47
0.0
47
1.0
290
0
70
0
71
2
292
1
293
0
294
0
295
1
296
0
11
0.0
21
0.0
31
0.0
"""


def test_lofted_surface_write_dxf():
    entity = LoftedSurface.from_text(LOFTEDSURFACE, doc=MockDoc())
    result = special_export(entity, dxfversion=DXF2010)
    expected = basic_tags_from_text(LOFTEDSURFACE)
    assert result == expected


LOFTEDSURFACE = """0
LOFTEDSURFACE
5
0
330
0
100
AcDbEntity
8
0
100
AcDbModelerGeometry
70
1
100
AcDbSurface
71
0
72
0
100
AcDbLoftedSurface
40
1.0
40
0.0
40
0.0
40
0.0
40
0.0
40
1.0
40
0.0
40
0.0
40
0.0
40
0.0
40
1.0
40
0.0
40
0.0
40
0.0
40
0.0
40
1.0
70
0
41
0.0
42
0.0
43
0.0
44
0.0
290
0
291
1
292
1
293
1
294
0
295
0
296
0
297
1
"""


def test_revolved_surface_write_dxf():
    entity = RevolvedSurface.from_text(REVOLVEDSURFACE, doc=MockDoc())
    result = special_export(entity, dxfversion=DXF2010)
    expected = basic_tags_from_text(REVOLVEDSURFACE)
    assert result == expected


REVOLVEDSURFACE = """0
REVOLVEDSURFACE
5
0
330
0
100
AcDbEntity
8
0
100
AcDbModelerGeometry
70
1
100
AcDbSurface
71
0
72
0
100
AcDbRevolvedSurface
90
36
10
0.0
20
0.0
30
0.0
11
0.0
21
0.0
31
1.0
40
0.0
41
0.0
42
1.0
42
0.0
42
0.0
42
0.0
42
0.0
42
1.0
42
0.0
42
0.0
42
0.0
42
0.0
42
1.0
42
0.0
42
0.0
42
0.0
42
0.0
42
1.0
43
0.0
44
0.0
45
0.0
46
0.0
290
0
291
0
"""


def test_swept_surface_write_dxf():
    entity = SweptSurface.from_text(SWEPTSURFACE, doc=MockDoc())
    result = special_export(entity, dxfversion=DXF2010)
    expected = basic_tags_from_text(SWEPTSURFACE)
    assert result == expected


SWEPTSURFACE = """0
SWEPTSURFACE
5
0
330
0
100
AcDbEntity
8
0
100
AcDbModelerGeometry
70
1
100
AcDbSurface
71
0
72
0
100
AcDbSweptSurface
90
36
91
36
40
1.0
40
0.0
40
0.0
40
0.0
40
0.0
40
1.0
40
0.0
40
0.0
40
0.0
40
0.0
40
1.0
40
0.0
40
0.0
40
0.0
40
0.0
40
1.0
41
1.0
41
0.0
41
0.0
41
0.0
41
0.0
41
1.0
41
0.0
41
0.0
41
0.0
41
0.0
41
1.0
41
0.0
41
0.0
41
0.0
41
0.0
41
1.0
42
0.0
43
0.0
44
0.0
45
0.0
48
1.0
49
0.0
46
1.0
46
0.0
46
0.0
46
0.0
46
0.0
46
1.0
46
0.0
46
0.0
46
0.0
46
0.0
46
1.0
46
0.0
46
0.0
46
0.0
46
0.0
46
1.0
47
1.0
47
0.0
47
0.0
47
0.0
47
0.0
47
1.0
47
0.0
47
0.0
47
0.0
47
0.0
47
1.0
47
0.0
47
0.0
47
0.0
47
0.0
47
1.0
290
0
70
1
71
2
292
0
293
0
294
1
295
1
296
1
11
0.0
21
0.0
31
0.0
"""

if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2018-2019 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf


@pytest.fixture(scope="module")
def dxf12():
    dwg = ezdxf.new("R12", setup="all")
    return dwg


def test_dimstyle_standard_exist(dxf12):
    assert "EZDXF" in dxf12.dimstyles


def test_dimstyle_override(dxf12):
    msp = dxf12.modelspace()
    dimstyle = msp.add_linear_dim(
        base=(3, 2, 0),
        p1=(0, 0, 0),
        p2=(3, 0, 0),
        dxfattribs={
            "dimstyle": "EZDXF",
        },
    )
    dimline = dimstyle.dimension
    assert dimline.dxf.dimstyle == "EZDXF"

    preset = {
        "dimtxsty": "TEST",  # virtual attribute - 'dimtxsty_handle' stores the text style handle
        "dimexe": 0.777,
        "dimblk": ezdxf.ARROWS.dot_blank,
    }
    dimstyle.update(preset)
    assert dimstyle["dimtxsty"] == "TEST"
    assert dimstyle["dimexe"] == 0.777

    assert dimstyle["invalid"] is None

    # ignores invalid attributes
    dimstyle.update({"invalid": 7})
    assert dimstyle["invalid"] == 7
    dstyle_orig = dimstyle.get_dstyle_dict()
    assert len(dstyle_orig) == 0

    dimstyle.commit()
    # check group code 5 handling for 'dimblk'
    data = dimline.get_xdata_list("ACAD", "DSTYLE")
    for tag in data:
        if tag.value == ezdxf.ARROWS.dot_blank:
            assert (
                tag.code == 1000
            ), "Despite group code 5, 'dimblk' should be treated as string, not as handle"

    dstyle = dimstyle.get_dstyle_dict()
    assert dstyle["dimexe"] == 0.777
    # unsupported DXF DimStyle attributes are not stored in dstyle
    assert "dimtxsty" not in dstyle, "dimtxsty is not a DXF12 attribute"


def test_dimstyle_override_arrows(dxf12):
    msp = dxf12.modelspace()
    preset = {
        "dimblk": "XYZ",
        "dimblk1": "ABC",
        "dimblk2": "DEF",
        "dimldrblk": "ZZZLDR",  # not supported by DXF12, but used by ezdxf for rendering
    }

    dimstyle = msp.add_linear_dim(
        base=(3, 2, 0),
        p1=(0, 0, 0),
        p2=(3, 0, 0),
        dimstyle="EZDXF",
        override=preset,
    )

    assert dimstyle["dimblk"] == "XYZ"
    assert dimstyle["dimblk1"] == "ABC"
    assert dimstyle["dimblk2"] == "DEF"
    assert dimstyle["dimldrblk"] == "ZZZLDR"

    dstyle_orig = dimstyle.get_dstyle_dict()
    assert len(dstyle_orig) == 0

    dimstyle.commit()
    dstyle = dimstyle.get_dstyle_dict()
    assert dstyle["dimblk"] == "XYZ"
    assert dstyle["dimblk1"] == "ABC"
    assert dstyle["dimblk2"] == "DEF"
    assert "dimldrblk" not in dstyle, "not supported by DXF12"

    dimstyle.set_arrows(blk="BLK", blk1="B1", blk2="B2", ldrblk="LBLK")
    assert dimstyle["dimblk"] == "BLK"
    assert dimstyle["dimblk1"] == "B1"
    assert dimstyle["dimblk2"] == "B2"
    assert (
        dimstyle["dimldrblk"] == "LBLK"
    )  # not supported by DXF12, but used by ezdxf for rendering

    dimstyle.commit()
    dstyle = dimstyle.get_dstyle_dict()
    assert dstyle["dimblk"] == "BLK"
    assert dstyle["dimblk1"] == "B1"
    assert dstyle["dimblk2"] == "B2"
    assert "dimnldrblk" not in dstyle, "not supported by DXF12"


def test_dimstyle_override_linetypes(dxf12):
    msp = dxf12.modelspace()
    preset = {
        "dimltype": "DOT",
        "dimltex1": "DOT2",
        "dimltex2": "DOTX2",
    }
    dimstyle = msp.add_linear_dim(
        base=(3, 2, 0),
        p1=(0, 0, 0),
        p2=(3, 0, 0),
        dimstyle="EZDXF",
        override=preset,
    )
    assert dimstyle["dimltype"] == "DOT"
    assert dimstyle["dimltex1"] == "DOT2"
    assert dimstyle["dimltex2"] == "DOTX2"

    dimstyle.commit()
    dstyle = dimstyle.get_dstyle_dict()
    assert "dimltype" not in dstyle, "not supported by DXF12"
    assert "dimltype_handle" not in dstyle, "not supported by DXF12"
    assert "dimltex1" not in dstyle, "not supported by DXF12"
    assert "dimltex1_handle" not in dstyle, "not supported by DXF12"
    assert "dimltex2" not in dstyle, "not supported by DXF12"
    assert "dimltex2_handle" not in dstyle, "not supported by DXF12"


def test_horizontal_dimline(dxf12):
    msp = dxf12.modelspace()
    dimstyle = msp.add_linear_dim(
        base=(3, 2, 0),
        p1=(0, 0, 0),
        p2=(3, 0, 0),
    )
    dimline = dimstyle.dimension
    assert dimline.dxf.dimstyle == "EZDXF"

    dimstyle.render()
    block_name = dimline.dxf.geometry
    assert block_name.startswith("*D")

    # shortcut for: block = dxf12.blocks.get(block_name)
    block = dimline.get_geometry_block()
    assert len(list(block.query("TEXT"))) == 1
    assert len(list(block.query("INSERT"))) == 2
    assert (
        len(list(block.query("LINE"))) == 3
    )  # dimension line + 2 extension lines
    assert len(list(block.query("POINT"))) == 3  # def points



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf

from ezdxf.entities.insert import Insert
from ezdxf.lldxf.const import DXF12, DXF2000
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.math import Matrix44
from ezdxf.entities import factory
from ezdxf.protocols import SupportsVirtualEntities, query_virtual_entities

ENTITY_R12 = """0
INSERT
5
0
8
0
2
BLOCKNAME
10
0.0
20
0.0
30
0.0
41
1.0
42
1.0
43
1.0
50
0.0
"""

ENTITY_R2000 = """0
INSERT
5
0
330
0
100
AcDbEntity
8
0
100
AcDbBlockReference
2
BLOCKNAME
10
0.0
20
0.0
30
0.0
41
1.0
42
1.0
43
1.0
50
0.0
"""


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


@pytest.fixture(scope="module")
def layout(doc):
    return doc.modelspace()


@pytest.fixture(params=[ENTITY_R12, ENTITY_R2000])
def entity(request):
    return Insert.from_text(request.param)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "INSERT" in ENTITY_CLASSES


def test_default_constructor():
    insert = Insert()
    assert insert.dxftype() == "INSERT"
    assert insert.is_virtual
    assert insert.seqend is None, "SEQEND must not exist"


def test_new_constructor():
    insert = Insert.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "insert": (1, 2, 3),
        },
    )
    assert insert.is_virtual is True, "Has no assigned document"
    assert insert.dxf.layer == "0"
    assert insert.dxf.color == 7
    assert insert.dxf.linetype == "BYLAYER"
    assert insert.dxf.insert == (1, 2, 3)
    assert insert.dxf.insert.x == 1, "is not Vec3 compatible"
    assert insert.dxf.insert.y == 2, "is not Vec3 compatible"
    assert insert.dxf.insert.z == 3, "is not Vec3 compatible"
    assert insert.has_scaling is False
    assert insert.has_uniform_scaling is True
    # can set DXF R2007 value
    insert.dxf.shadow_mode = 1
    assert insert.dxf.shadow_mode == 1


def test_has_scaling():
    entity = Insert.new(handle="ABBA", owner="0", dxfattribs={"xscale": 2})
    assert entity.has_scaling is True
    assert entity.has_uniform_scaling is False
    entity = Insert.new(handle="ABBA", owner="0", dxfattribs={"yscale": 2})
    assert entity.has_scaling is True
    assert entity.has_uniform_scaling is False
    entity = Insert.new(handle="ABBA", owner="0", dxfattribs={"zscale": 2})
    assert entity.has_scaling is True
    assert entity.has_uniform_scaling is False

    # reflections are under control, so (-2, 2, 2) is a uniform scaling
    entity = Insert.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "xscale": -2,
            "yscale": 2,
            "zscale": 2,
        },
    )
    assert entity.has_uniform_scaling is True


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.insert == (0, 0, 0)
    assert entity.has_scaling is False


@pytest.mark.parametrize("txt,ver", [(ENTITY_R2000, DXF2000), (ENTITY_R12, DXF12)])
def test_write_dxf(txt, ver):
    expected = basic_tags_from_text(txt)
    vertex = Insert.from_text(txt)
    collector = TagCollector(dxfversion=ver, optional=True)
    vertex.export_dxf(collector)
    assert collector.tags == expected

    collector2 = TagCollector(dxfversion=ver, optional=False)
    vertex.export_dxf(collector2)
    assert collector.has_all_tags(collector2)


def test_add_attribs():
    insert = Insert()
    assert insert.attribs_follow is False
    insert.add_attrib("T1", "value1", (0, 0))
    assert len(insert.attribs) == 1
    assert insert.attribs_follow is True


def test_get_block(doc):
    blk = doc.blocks.new("TEST-2020-02-29")
    msp = doc.modelspace()
    insert = msp.add_blockref("TEST-2020-02-29", (0, 0))
    assert insert.block() is blk
    insert = msp.add_blockref("NOT-EXIST-2020-02-29", (0, 0))
    assert insert.block() is None
    insert = Insert()
    assert insert.block() is None


def test_clone_with_insert():
    # difference of clone() to copy_entity() is:
    # - clone returns and unassigned entity without handle, owner or reactors
    # - copy_entity clones the entity and assigns the new entity to the same owner as the source and adds the entity
    #   and it linked entities (ATTRIB & VERTEX) to the entity database, but does not adding entity to a layout, setting
    #   owner tag is not enough to assign an entity to a layout, use Layout.add_entity()
    insert = Insert()
    insert.add_attrib("T1", "value1", (0, 0))
    clone = insert.copy()
    assert clone.dxf.handle is None
    assert clone.dxf.owner is None
    assert len(clone.attribs) == 1
    attrib = clone.attribs[0]
    assert attrib.dxf.handle is None
    assert attrib.dxf.tag == "T1"
    assert attrib.dxf.text == "value1"
    # change cloned attrib
    attrib.dxf.tag = "T2"
    attrib.dxf.text = "value2"
    # source attrib is unchanged
    assert insert.attribs[0].dxf.tag == "T1"
    assert insert.attribs[0].dxf.text == "value1"


def test_export_without_sub_entities_to_dxf(doc):
    from ezdxf.lldxf.tagwriter import TagCollector

    blk = doc.blocks.new("INSERT_WITHOUT_ATTRIBS")
    blk.add_blockref("TEST", (0, 0))
    writer = TagCollector()
    blk.entity_space.export_dxf(tagwriter=writer)
    structure_tags = [tag for tag in writer.tags if tag[0] == 0]
    assert len(structure_tags) == 1
    assert structure_tags[0] == (0, "INSERT")


def test_export_with_sub_entities_to_dxf(doc):
    from ezdxf.lldxf.tagwriter import TagCollector

    blk = doc.blocks.new("INSERT_WITH_ATTRIBS")
    insert = blk.add_blockref("TEST", (0, 0))
    insert.add_attrib("TAG", "TEXT", (0, 0))
    writer = TagCollector()
    blk.entity_space.export_dxf(tagwriter=writer)
    structure_tags = [tag for tag in writer.tags if tag[0] == 0]
    assert structure_tags[0] == (0, "INSERT")
    assert structure_tags[1] == (0, "ATTRIB")
    assert structure_tags[2] == (0, "SEQEND")


def test_copy_with_insert(doc):
    msp = doc.modelspace()
    msp_count = len(msp)
    db = doc.entitydb
    db_len = len(doc.entitydb)

    insert = msp.add_blockref("Test", insert=(0, 0))
    assert (
        insert.seqend.dxf.owner == insert.dxf.handle
    ), "INSERT is owner of the SEQEND entity"
    assert insert.seqend.dxf.handle is not None
    assert insert.seqend.dxf.handle in doc.entitydb

    insert.add_attrib("T1", "value1")

    # linked attribs not stored in the entity space
    assert len(msp) == msp_count + 1
    # added INSERT + SEQEND
    assert len(db) == db_len + 3, "New ATTRIBS automatically stored in db"

    copy = doc.entitydb.duplicate_entity(insert)

    # not duplicated in entity space
    assert len(msp) == msp_count + 1
    # duplicated in entity database (2x SEQEND)
    assert len(doc.entitydb) == db_len + 6

    # get 1. paperspace in tab order
    psp = doc.layout()
    psp.add_entity(copy)
    assert len(psp) == 1

    assert copy.dxf.handle is not None
    assert copy.dxf.handle != insert.dxf.handle
    assert copy.dxf.owner == psp.layout_key
    assert (
        copy.attribs[0].dxf.owner == copy.dxf.handle
    ), "INSERT is owner of ATTRIB entities"


def test_matrix44_no_transform():
    insert = Insert.new(handle="ABBA", owner="0")
    m = insert.matrix44()
    assert m.transform((0, 0, 0)) == (0, 0, 0)
    assert m.transform_direction((1, 0, 0)) == (1, 0, 0)


def test_matrix44_insert():
    insert = Insert.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "insert": (1, 2, 3),
        },
    )
    m = insert.matrix44()
    assert m.transform((0, 0, 0)) == (1, 2, 3)
    assert m.transform_direction((1, 0, 0)) == (1, 0, 0)


def test_matrix44_insert_and_base_point(doc):
    doc.blocks.new("Matrix44_001", base_point=(2, 2, 2))
    insert = doc.modelspace().add_blockref("Matrix44_001", insert=(1, 2, 3))
    m = insert.matrix44()
    assert m.transform((0, 0, 0)) == (-1, 0, 1)
    assert m.transform_direction((1, 0, 0)) == (1, 0, 0)


def test_matrix44_rotation():
    insert = Insert.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "insert": (0, 0, 0),
            "rotation": 90,
        },
    )
    m = insert.matrix44()
    for v1, v2 in zip(
        m.transform_vertices([(1, 0, 0), (0, 0, 1)]), [(0, 1, 0), (0, 0, 1)]
    ):
        assert v1.isclose(v2)
    assert m.transform_direction((1, 0, 0)).isclose((0, 1, 0))


def test_matrix44_scaled():
    insert = Insert.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "xscale": 2,
            "yscale": 3,
            "zscale": 4,
        },
    )
    m = insert.matrix44()
    assert m.transform((1, 1, 1)) == (2, 3, 4)
    assert m.transform_direction((1, 0, 0)) == (
        2,
        0,
        0,
    ), "scaling has to be applied for directions"


def test_matrix44_direction():
    insert = Insert.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "insert": (1, 2, 3),
            "xscale": 2,
        },
    )
    m = insert.matrix44()
    assert m.transform((1, 0, 0)) == (3, 2, 3)
    assert m.transform_direction((1, 0, 0)) == (
        2,
        0,
        0,
    ), "only scaling has to be applied for directions"


def test_insert_transform_interface():
    insert = Insert()
    insert.dxf.insert = (1, 0, 0)

    insert.transform(Matrix44.translate(1, 2, 3))
    assert insert.dxf.insert == (2, 2, 3)

    # optimized translate implementation
    insert.translate(-1, -2, -3)
    assert insert.dxf.insert == (1, 0, 0)


def test_insert_scaling():
    # Insert.transform() changes the extrusion vector
    # sign of scaling factors depend from new extrusion vector
    # just compare absolute values
    insert = Insert()
    insert.dxf.insert = (0, 0, 0)

    insert.scale(2, 3, 4)
    assert abs(insert.dxf.xscale) == 2
    assert abs(insert.dxf.yscale) == 3
    assert abs(insert.dxf.zscale) == 4

    insert.scale(-1, 1, 1)
    assert abs(insert.dxf.xscale) == 2
    assert abs(insert.dxf.yscale) == 3
    assert abs(insert.dxf.zscale) == 4

    insert.scale(-1, -1, 1)
    assert abs(insert.dxf.xscale) == 2
    assert abs(insert.dxf.yscale) == 3
    assert abs(insert.dxf.zscale) == 4

    insert.scale(-2, -2, -2)
    assert abs(insert.dxf.xscale) == 4
    assert abs(insert.dxf.yscale) == 6
    assert abs(insert.dxf.zscale) == 8


def test_add_virtual_insert_with_attribs_to_layout(doc):
    doc.blocks.new("TestAddVirtualInsert")
    msp = doc.modelspace()
    insert = Insert.new(dxfattribs={"name": "TestAddVirtualInsert"})
    insert.add_attrib("TAG", "TEXT", (0, 0))
    msp.add_entity(insert)

    assert factory.is_bound(insert, doc) is True
    assert (
        factory.is_bound(insert.seqend, doc) is True
    ), "SEQEND must be bound to document"

    assert insert.attribs_follow is True
    assert (
        factory.is_bound(insert.attribs[0], doc) is True
    ), "ATTRIB must be bound to document"


class TestSupportsVirtualEntitiesProtocol:
    @pytest.fixture(scope="class")
    def doc(self):
        doc = ezdxf.new()
        blk = doc.blocks.new("POINT")
        blk.add_point(location=(0, 0))
        return doc

    @pytest.fixture
    def insert(self, doc):
        msp = doc.modelspace()
        return msp.add_blockref("POINT", (0, 0))

    def test_supports_virtual_entities_protocol(self, insert):
        assert isinstance(insert, SupportsVirtualEntities) is True

    def test_query_virtual_entities(self, insert):
        entities = query_virtual_entities(insert)
        assert len(entities) == 1
        assert entities[0].dxftype() == "POINT"


MALFORMED_INSERT = """0
INSERT
2
BLOCKNAME
5
0
62
7
330
0
6
LT_EZDXF
8
LY_EZDXF
100
AcDbEntity
100
AcDbArc
2
BLOCKNAME
10
1.0
20
2.0
30
3.0
41
1.0
42
1.0
43
1.0
100
AcDbBlockReference
50
0.0
"""


def test_load_malformed_insert():
    insert = Insert.from_text(MALFORMED_INSERT)
    assert insert.dxf.layer == "LY_EZDXF"
    assert insert.dxf.linetype == "LT_EZDXF"
    assert insert.dxf.color == 7
    assert insert.dxf.insert.isclose((1, 2, 3))
    assert insert.dxf.name == "BLOCKNAME"



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations

import string
from ezdxf import const
from ezdxf.lldxf.types import dxftag
from ezdxf.entities import XData, DXFEntity, is_graphic_entity
from ezdxf.document import Drawing
from ezdxf.sections.table import Table

__all__ = ["make_acad_compatible", "translate_names", "clean", "R12NameTranslator"]


def make_acad_compatible(doc: Drawing) -> None:
    """Apply all DXF R12 requirements, so Autodesk products will load the document."""
    if doc.dxfversion != const.DXF12:
        raise const.DXFVersionError(
            f"expected DXF document version R12, got: {doc.acad_release}"
        )
    clean(doc)
    translate_names(doc)


def translate_names(doc: Drawing) -> None:
    r"""Translate table and block names into strict DXF R12 names.

    ACAD Releases upto 14 limit names to 31 characters in length and all names are
    uppercase.  Names can include the letters A to Z, the numerals 0 to 9, and the
    special characters, dollar sign ($), underscore (_), hyphen (-) and the
    asterix (\*) as first character for special names like anonymous blocks.

    Most applications do not care about that and work fine with longer names and
    any characters used in names for some exceptions, but of course Autodesk
    applications are very picky about that.

    .. note::

        This is a destructive process and modifies the internals of the DXF document.

    """
    if doc.dxfversion != const.DXF12:
        raise const.DXFVersionError(
            f"expected DXF document version R12, got: {doc.acad_release}"
        )
    _R12StrictRename(doc).execute()


def clean(doc: Drawing) -> None:
    """Removes all features that are not supported for DXF R12 by Autodesk products."""
    if doc.dxfversion != const.DXF12:
        raise const.DXFVersionError(
            f"expected DXF document version R12, got: {doc.acad_release}"
        )
    _remove_table_xdata(doc.appids)
    _remove_table_xdata(doc.linetypes)
    _remove_table_xdata(doc.layers)
    _remove_table_xdata(doc.styles)
    _remove_table_xdata(doc.dimstyles)
    _remove_table_xdata(doc.ucs)
    _remove_table_xdata(doc.views)
    _remove_table_xdata(doc.viewports)
    _remove_legacy_blocks(doc)


def _remove_table_xdata(table: Table) -> None:
    """Autodesk products do not accept XDATA in table entries for DXF R12."""
    for entry in list(table):
        entry.xdata = None


def _remove_legacy_blocks(doc: Drawing) -> None:
    """Due to bad conversion some DXF files contain after loading the blocks
    "$MODEL_SPACE" and "$PAPER_SPACE". This function removes these empty blocks,
    because they will clash with the translated layout names of "*Model_Space" and
    "*Paper_Space".
    """
    for name in ("$MODEL_SPACE", "$PAPER_SPACE"):
        try:
            doc.blocks.delete_block(name, safe=False)
        except const.DXFKeyError:
            pass


class R12NameTranslator:
    r"""Translate table and block names into strict DXF R12 names.

    ACAD Releases upto 14 limit names to 31 characters in length and all names are
    uppercase.  Names can include the letters A to Z, the numerals 0 to 9, and the
    special characters, dollar sign ($), underscore (_), hyphen (-) and the
    asterix (\*) as first character for special names like anonymous blocks.

    """

    VALID_R12_NAME_CHARS = set(string.ascii_uppercase + string.digits + "$_-")

    def __init__(self) -> None:
        self.translated_names: dict[str, str] = {}
        self.used_r12_names: set[str] = set()

    def reset(self) -> None:
        self.translated_names.clear()
        self.used_r12_names.clear()

    def translate(self, name: str) -> str:
        name = name.upper()
        r12_name = self.translated_names.get(name)
        if r12_name is None:
            r12_name = self._name_sanitizer(name, self.VALID_R12_NAME_CHARS)
            r12_name = self._get_unique_r12_name(r12_name)
            self.translated_names[name] = r12_name
        return r12_name

    def _get_unique_r12_name(self, name: str) -> str:
        name0 = name
        counter = 0
        while name in self.used_r12_names:
            ext = str(counter)
            name = name0[: (31 - len(ext))] + ext
            counter += 1
        self.used_r12_names.add(name)
        return name

    @staticmethod
    def _name_sanitizer(name: str, valid_chars: set[str]) -> str:
        # `name` has to be upper case!
        if not name:
            return ""
        new_name = "".join(
                (char if char in valid_chars else "_") for char in name[:31]
            )
        # special names like anonymous blocks or the "*ACTIVE" viewport configuration
        if name[0] == "*":
            return "*" + new_name[1:31]
        else:
            return new_name


COMMON_ATTRIBS = ["layer", "linetype", "style", "tag", "name", "dimstyle"]
DIMSTYLE_ATTRIBS = ["dimblk", "dimblk1", "dimblk2"]
LAYER_ATTRIBS = ["linetype"]
BLOCK_ATTRIBS = ["layer"]


class _R12StrictRename:
    def __init__(self, doc: Drawing) -> None:
        assert doc.dxfversion == const.DXF12, "expected DXF version R12"
        self.doc = doc
        self.translator = R12NameTranslator()

    def execute(self) -> None:
        self.process_tables()
        self.process_header_vars()
        self.process_entities()

    def process_tables(self) -> None:
        tables = self.doc.tables
        self.rename_table_entries(tables.appids)
        self.rename_table_entries(tables.linetypes)
        self.rename_table_entries(tables.layers)
        self.process_table_entries(tables.layers, LAYER_ATTRIBS)
        self.rename_table_entries(tables.styles)
        self.rename_table_entries(tables.dimstyles)
        self.process_table_entries(tables.dimstyles, DIMSTYLE_ATTRIBS)
        self.rename_table_entries(tables.ucs)
        self.rename_table_entries(tables.views)
        self.rename_vports()
        self.rename_block_layouts()
        self.process_blocks()

    def rename_table_entries(self, table: Table) -> None:
        translate = self.translator.translate
        for entry in list(table):
            name = entry.dxf.name
            entry.dxf.name = translate(name)
            table.replace(name, entry)
            # XDATA for table entries is not accepted by Autodesk products for R12,
            # but for consistency a translation is applied
            if entry.xdata:
                self.translate_xdata(entry.xdata)

    def rename_vports(self):
        translate = self.translator.translate
        for config in list(self.doc.viewports.entries.values()):
            if not config:  # multiple entries in a sequence
                continue
            old_name = config[0].dxf.name
            new_name = translate(old_name)
            for entry in config:
                entry.dxf.name = new_name
            self.doc.viewports.replace(old_name, config)

    def process_table_entries(self, table: Table, attribute_names: list[str]) -> None:
        for entry in table:
            self.translate_entity_attributes(entry, attribute_names)

    def rename_block_layouts(self) -> None:
        translate = self.translator.translate
        blocks = self.doc.blocks
        for name in blocks.block_names():
            blocks.rename_block(name, translate(name))

    def process_blocks(self):
        for block_record in self.doc.block_records:
            block = block_record.block
            self.translate_entity_attributes(block, BLOCK_ATTRIBS)
            if block.xdata:
                self.translate_xdata(block.xdata)

    def process_header_vars(self) -> None:
        header = self.doc.header
        translate = self.translator.translate

        for key in (
            "$CELTYPE",
            "$CLAYER",
            "$DIMBLK",
            "$DIMBLK1",
            "$DIMBLK2",
            "$DIMSTYLE",
            "$UCSNAME",
            "$PUCSNAME",
            "$TEXTSTYLE",
        ):
            value = self.doc.header.get(key)
            if value:
                header[key] = translate(value)

    def process_entities(self) -> None:
        for entity in self.doc.entitydb.values():
            if not is_graphic_entity(entity):
                continue
            if entity.MIN_DXF_VERSION_FOR_EXPORT > const.DXF12:
                continue
            if entity.xdata:
                self.translate_xdata(entity.xdata)
            self.translate_entity_attributes(entity, COMMON_ATTRIBS)

    def translate_entity_attributes(
        self, entity: DXFEntity, attribute_names: list[str]
    ) -> None:
        translate = self.translator.translate
        for attrib_name in attribute_names:
            if not entity.dxf.hasattr(attrib_name):
                continue
            name = entity.dxf.get(attrib_name)
            if name:
                entity.dxf.set(attrib_name, translate(name))

    def translate_xdata(self, xdata: XData) -> None:
        translate = self.translator.translate
        for tags in xdata.data.values():
            for index, (code, value) in enumerate(tags):
                # 1001: APPID
                # 1003: layer name
                if code == 1001 or code == 1003:
                    tags[index] = dxftag(code, translate(value))
        xdata.update_keys()



# Copyright (c) 2021, Manfred Moitzi
# License: MIT License
#
# Purpose:
# Flips an inverted OCS defined by extrusion vector (0, 0, -1) into a WCS
# aligned OCS defined by extrusion vector (0, 0, 1).
#
# This simplifies 2D entity processing for ezdxf users and creates DXF
# output for 3rd party DXF libraries which ignore the existence of the OCS.
#
# Warning:
# The WCS representation of OCS entities with flipped extrusion vector
# is not 100% identical to the source entity, curve orientation and vertex order
# may change. A mirrored text represented by an extrusion vector (0, 0, -1)
# cannot represented by an extrusion vector (0, 0, 1), therefore this CANNOT
# work for text entities or entities including text:
# TEXT, ATTRIB, ATTDEF, MTEXT, DIMENSION, LEADER, MLEADER
from __future__ import annotations
from typing import Iterable, Sequence
from functools import singledispatch
import math
from ezdxf.math import Z_AXIS, Vec3, Vec2
from ezdxf.entities import (
    DXFGraphic,
    DXFNamespace,
    Circle,
    Arc,
    Ellipse,
    Solid,
    Insert,
    LWPolyline,
    Polyline,
)
from ezdxf.entities.polygon import DXFPolygon
from ezdxf.entities.boundary_paths import (
    PolylinePath,
    EdgePath,
    LineEdge,
    ArcEdge,
    EllipseEdge,
    SplineEdge,
)
from ezdxf.lldxf import const

__all__ = ["upright", "upright_all"]


def upright(entity: DXFGraphic) -> None:
    """Flips an inverted :ref:`OCS` defined by extrusion vector (0, 0, -1) into
    a :ref:`WCS` aligned :ref:`OCS` defined by extrusion vector (0, 0, 1).
    DXF entities with other extrusion vectors and unsupported DXF entities will
    be silently ignored. For more information about the limitations read the
    documentation of the :mod:`ezdxf.upright` module.

    """
    if not (
        isinstance(entity, DXFGraphic)
        and entity.is_alive
        and entity.dxf.hasattr("extrusion")
    ):
        return
    extrusion = Vec3(entity.dxf.extrusion).normalize()
    if extrusion.isclose(FLIPPED_Z_AXIS):
        _flip_dxf_graphic(entity)


def upright_all(entities: Iterable[DXFGraphic]) -> None:
    """Call function :func:`upright` for all DXF entities in iterable
    `entities`::

        upright_all(doc.modelspace())

    """
    for e in entities:
        upright(e)


FLIPPED_Z_AXIS = -Z_AXIS


def _flip_deg_angle(angle: float) -> float:
    return (180.0 if angle >= 0.0 else -180.0) - angle


def _flip_rad_angle(angle: float) -> float:
    return (math.pi if angle >= 0.0 else -math.pi) - angle


def _flip_vertex(vertex: Vec3) -> Vec3:
    return Vec3(-vertex.x, vertex.y, -vertex.z)


def _flip_2d_vertex(vertex: Vec2) -> Vec2:
    return Vec2(-vertex.x, vertex.y)


def _flip_existing_vertex(dxf: DXFNamespace, name: str) -> None:
    if dxf.hasattr(name):
        vertex = _flip_vertex(dxf.get(name))
        dxf.set(name, vertex)


def _flip_thickness(dxf: DXFNamespace) -> None:
    if dxf.hasattr("thickness"):
        dxf.thickness = -dxf.thickness


def _flip_elevation(dxf: DXFNamespace) -> None:
    if dxf.hasattr("elevation"):
        # works with float (LWPOLYLINE) and Vec3 (POLYLINE)
        dxf.elevation = -dxf.elevation


@singledispatch
def _flip_dxf_graphic(entity: DXFGraphic) -> None:
    pass  # ignore unsupported types on purpose


@_flip_dxf_graphic.register(Circle)
def _flip_circle(circle: Circle) -> None:
    dxf = circle.dxf
    _flip_existing_vertex(dxf, "center")
    _flip_thickness(dxf)
    dxf.discard("extrusion")


@_flip_dxf_graphic.register(Arc)
def _flip_arc(arc: Arc) -> None:
    _flip_circle(arc)
    dxf = arc.dxf
    end_angle = dxf.end_angle
    dxf.end_angle = _flip_deg_angle(dxf.start_angle)
    dxf.start_angle = _flip_deg_angle(end_angle)


# SOLID and TRACE - but not 3DFACE!
@_flip_dxf_graphic.register(Solid)
def _flip_solid(solid: Solid) -> None:
    dxf = solid.dxf
    for name in const.VERTEXNAMES:
        _flip_existing_vertex(dxf, name)
    _flip_thickness(dxf)
    dxf.discard("extrusion")


@_flip_dxf_graphic.register(Insert)
def _flip_insert(insert: Insert) -> None:
    # see exploration/upright_insert.py
    dxf = insert.dxf
    _flip_existing_vertex(dxf, "insert")
    dxf.rotation = -dxf.rotation
    dxf.xscale = -dxf.xscale
    dxf.zscale = -dxf.zscale
    # Attached attributes cannot be flipped!
    dxf.discard("extrusion")


@_flip_dxf_graphic.register(Ellipse)
def _flip_ellipse(ellipse: Ellipse) -> None:
    # ELLIPSE is a WCS entity!
    # just process start- and end params
    dxf = ellipse.dxf
    end_param = -dxf.end_param
    dxf.end_param = -dxf.start_param
    dxf.start_param = end_param
    dxf.discard("extrusion")


@_flip_dxf_graphic.register(LWPolyline)
def _flip_lwpolyline(polyline: LWPolyline) -> None:
    flipped_points: list[Sequence[float]] = []
    for x, y, start_width, end_width, bulge in polyline.lwpoints:
        bulge = -bulge
        v = _flip_2d_vertex(Vec2(x, y))
        flipped_points.append((v.x, v.y, start_width, end_width, bulge))
    polyline.set_points(flipped_points, format="xyseb")
    _finalize_polyline(polyline.dxf)


def _finalize_polyline(dxf: DXFNamespace):
    _flip_thickness(dxf)
    _flip_elevation(dxf)
    dxf.discard("extrusion")


@_flip_dxf_graphic.register(Polyline)
def _flip_polyline2d(polyline: Polyline) -> None:
    if not polyline.is_2d_polyline:
        return  # ignore silently
    for vertex in polyline.vertices:
        dxf = vertex.dxf
        _flip_existing_vertex(dxf, "location")
        if dxf.hasattr("bulge"):
            dxf.bulge = -dxf.bulge
    _finalize_polyline(polyline.dxf)


# HATCH and MPOLYGON
@_flip_dxf_graphic.register(DXFPolygon)
def _flip_polygon(polygon: DXFPolygon) -> None:
    for p in polygon.paths:
        _flip_boundary_path(p)
    _flip_elevation(polygon.dxf)
    polygon.dxf.discard("extrusion")


@singledispatch
def _flip_boundary_path(path) -> None:
    raise TypeError(f"unsupported path type: {path!r}")


@_flip_boundary_path.register(PolylinePath)
def _flip_polyline_path(polyline: PolylinePath) -> None:
    flipped_vertices: list[tuple[float, float, float]] = []
    for x, y, bulge in polyline.vertices:
        bulge = -bulge
        v = _flip_2d_vertex(Vec2(x, y))
        flipped_vertices.append((v.x, v.y, bulge))
    polyline.vertices = flipped_vertices


@_flip_boundary_path.register(EdgePath)
def _flip_edge_path(edges: EdgePath) -> None:
    # see exploration/upright_hatch.py
    for edge in edges:
        _flip_edge(edge)


@singledispatch
def _flip_edge(edge) -> None:
    raise TypeError(f"unsupported edge type: {edge!r}")


@_flip_edge.register(LineEdge)
def _flip_line_edge(edge: LineEdge) -> None:
    edge.start = _flip_2d_vertex(edge.start)
    edge.end = _flip_2d_vertex(edge.end)


@_flip_edge.register(ArcEdge)
def _flip_arc_edge(edge: ArcEdge) -> None:
    edge.center = _flip_2d_vertex(edge.center)
    # Start- and end angles are always stored in counter-clockwise orientation!
    end_angle = edge.end_angle
    edge.end_angle = _flip_deg_angle(edge.start_angle)
    edge.start_angle = _flip_deg_angle(end_angle)
    edge.ccw = not edge.ccw


@_flip_edge.register(EllipseEdge)
def _flip_ellipse_edge(edge: EllipseEdge) -> None:
    edge.center = _flip_2d_vertex(edge.center)
    edge.major_axis = _flip_2d_vertex(edge.major_axis)
    # Ellipse params as angles in degrees - not radians!
    # Do not exchange start- and end angles!
    # see exploration/upright_hatch.py
    edge.start_angle = _flip_deg_angle(edge.start_angle)
    edge.end_angle = _flip_deg_angle(edge.end_angle)
    edge.ccw = not edge.ccw


@_flip_edge.register(SplineEdge)
def _flip_spline_edge(edge: SplineEdge) -> None:
    flip_2d_vertex = _flip_2d_vertex
    if edge.start_tangent is not None:
        edge.start_tangent = flip_2d_vertex(edge.start_tangent)
    if edge.end_tangent is not None:
        edge.end_tangent = flip_2d_vertex(edge.end_tangent)
    edge.control_points = [flip_2d_vertex(v) for v in edge.control_points]
    edge.fit_points = [flip_2d_vertex(v) for v in edge.fit_points]



# Copyright (c) 2010-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Iterable, Sequence
from functools import lru_cache
import math
import numpy as np

from ezdxf.math import Vec3, NULLVEC, Matrix44, UVec


__all__ = ["Bezier"]


"""

Bezier curves
=============

https://www.cl.cam.ac.uk/teaching/2000/AGraphHCI/SMEG/node3.html

A Bezier curve is a weighted sum of n+1 control points,  P0, P1, ..., Pn, where 
the weights are the Bernstein polynomials. 

The Bezier curve of order n+1 (degree n) has n+1 control points. These are the 
first three orders of Bezier curve definitions. 

(75) linear P(t) = (1-t)*P0 + t*P1
(76) quadratic P(t) = (1-t)^2*P0 + 2*(t-1)*t*P1 + t^2*P2
(77) cubic P(t) = (1-t)^3*P0 + 3*(1-t)^2*t*P1 + 3*(1-t)*t^2*P2 + t^3*P3

Ways of thinking about Bezier curves
------------------------------------

There are several useful ways in which you can think about Bezier curves. 
Here are the ones that I use.

Linear interpolation
~~~~~~~~~~~~~~~~~~~~

Equation (75) is obviously a linear interpolation between two points. Equation 
(76) can be rewritten as a linear interpolation between linear interpolations 
between points.

Weighted average
~~~~~~~~~~~~~~~~

A Bezier curve can be seen as a weighted average of all of its control points. 
Because all of the weights are positive, and because the weights sum to one, the 
Bezier curve is guaranteed to lie within the convex hull of its control points.
    
Refinement of the control polygon
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A Bezier curve can be seen as some sort of refinement of the polygon made by 
connecting its control points in order. The Bezier curve starts and ends at the 
two end points and its shape is determined by the relative positions of the n-1 
other control points, although it will generally not pass through these other 
control points. The tangent vectors at the start and end of the curve pass 
through the end point and the immediately adjacent point.

Continuity
----------

You should note that each Bezier curve is independent of any other Bezier curve. 
If we wish two Bezier curves to join with any type of continuity, then we must 
explicitly position the control points of the second curve so that they bear 
the appropriate relationship with the control points in the first curve.

Any Bezier curve is infinitely differentiable within itself, and is therefore 
continuous to any degree.

"""


class Bezier:
    """Generic `Bézier curve`_ of any degree.

    A `Bézier curve`_ is a parametric curve used in computer graphics and
    related fields. Bézier curves are used to model smooth curves that can be
    scaled indefinitely. "Paths", as they are commonly referred to in image
    manipulation programs, are combinations of linked Bézier curves.
    Paths are not bound by the limits of rasterized images and are intuitive to
    modify. (Source: Wikipedia)

    This is a generic implementation which works with any count of definition
    points greater than 2, but it is a simple and slow implementation. For more
    performance look at the specialized :class:`Bezier4P` and :class:`Bezier3P`
    classes.

    Objects are immutable.

    Args:
        defpoints: iterable of definition points as :class:`Vec3` compatible objects.

    """

    def __init__(self, defpoints: Iterable[UVec]):
        self._defpoints: Sequence[Vec3] = Vec3.tuple(defpoints)

    @property
    def control_points(self) -> Sequence[Vec3]:
        """Control points as tuple of :class:`Vec3` objects."""
        return self._defpoints

    def approximate(self, segments: int = 20) -> Iterable[Vec3]:
        """Approximates curve by vertices as :class:`Vec3` objects, vertices
        count = segments + 1.
        """
        return self.points(self.params(segments))

    def flattening(self, distance: float, segments: int = 4) -> Iterable[Vec3]:
        """Adaptive recursive flattening. The argument `segments` is the
        minimum count of approximation segments, if the distance from the center
        of the approximation segment to the curve is bigger than `distance` the
        segment will be subdivided.

        Args:
            distance: maximum distance from the center of the curve (Cn)
                to the center of the linear (C1) curve between two
                approximation points to determine if a segment should be
                subdivided.
            segments: minimum segment count

        """

        def subdiv(start_point, end_point, start_t: float, end_t: float):
            mid_t = (start_t + end_t) * 0.5
            mid_point = self.point(mid_t)
            chk_point = start_point.lerp(end_point)
            # center point point is faster than projecting mid point onto
            # vector start -> end:
            if chk_point.distance(mid_point) < distance:
                yield end_point
            else:
                yield from subdiv(start_point, mid_point, start_t, mid_t)
                yield from subdiv(mid_point, end_point, mid_t, end_t)

        dt = 1.0 / segments
        t0 = 0.0
        start_point = self._defpoints[0]
        yield start_point
        while t0 < 1.0:
            t1 = t0 + dt
            if math.isclose(t1, 1.0):
                end_point = self._defpoints[-1]
                t1 = 1.0
            else:
                end_point = self.point(t1)
            yield from subdiv(start_point, end_point, t0, t1)
            t0 = t1
            start_point = end_point

    def params(self, segments: int) -> Iterable[float]:
        """Yield evenly spaced parameters from 0 to 1 for given segment count."""
        yield from np.linspace(0.0, 1.0, segments + 1)

    def point(self, t: float) -> Vec3:
        """Returns a point for parameter `t` in range [0, 1] as :class:`Vec3`
        object.
        """
        if t < 0.0 or t > 1.0:
            raise ValueError("Parameter t not in range [0, 1]")
        if (1.0 - t) < 5e-6:
            t = 1.0
        point = NULLVEC
        pts = self._defpoints
        n = len(pts)

        for i in range(n):
            point += bernstein_basis(n - 1, i, t) * pts[i]
        return point

    def points(self, t: Iterable[float]) -> Iterable[Vec3]:
        """Yields multiple points for parameters in vector `t` as :class:`Vec3`
        objects. Parameters have to be in range [0, 1].
        """
        for u in t:
            yield self.point(u)

    def derivative(self, t: float) -> tuple[Vec3, Vec3, Vec3]:
        """Returns (point, 1st derivative, 2nd derivative) tuple for parameter `t`
        in range [0, 1] as :class:`Vec3` objects.
        """
        if t < 0.0 or t > 1.0:
            raise ValueError("Parameter t not in range [0, 1]")

        if (1.0 - t) < 5e-6:
            t = 1.0
        pts = self._defpoints
        n = len(pts)
        n0 = n - 1
        point = NULLVEC
        d1 = NULLVEC
        d2 = NULLVEC
        t2 = t * t
        n0_1 = n0 - 1
        if t == 0.0:
            d1 = n0 * (pts[1] - pts[0])
            d2 = n0 * n0_1 * (pts[0] - 2.0 * pts[1] + pts[2])
        for i in range(n):
            tmp_bas = bernstein_basis(n0, i, t)
            point += tmp_bas * pts[i]
            if 0.0 < t < 1.0:
                _1_t = 1.0 - t
                i_n0_t = i - n0 * t
                d1 += i_n0_t / (t * _1_t) * tmp_bas * pts[i]
                d2 += (
                    (i_n0_t * i_n0_t - n0 * t2 - i * (1.0 - 2.0 * t))
                    / (t2 * _1_t * _1_t)
                    * tmp_bas
                    * pts[i]
                )
            if t == 1.0:
                d1 = n0 * (pts[n0] - pts[n0_1])
                d2 = n0 * n0_1 * (pts[n0] - 2 * pts[n0_1] + pts[n0 - 2])
        return point, d1, d2

    def derivatives(
        self, t: Iterable[float]
    ) -> Iterable[tuple[Vec3, Vec3, Vec3]]:
        """Returns multiple (point, 1st derivative, 2nd derivative) tuples for
        parameter vector  `t` as :class:`Vec3` objects.
        Parameters in range [0, 1]
        """
        for u in t:
            yield self.derivative(u)

    def reverse(self) -> Bezier:
        """Returns a new Bèzier-curve with reversed control point order."""
        return Bezier(list(reversed(self.control_points)))

    def transform(self, m: Matrix44) -> Bezier:
        """General transformation interface, returns a new :class:`Bezier` curve.

        Args:
             m: 4x4 transformation matrix (:class:`ezdxf.math.Matrix44`)

        """
        defpoints = tuple(m.transform_vertices(self.control_points))
        return Bezier(defpoints)


def bernstein_basis(n: int, i: int, t: float) -> float:
    # handle the special cases to avoid domain problem with pow
    if t == 0.0 and i == 0:
        ti = 1.0
    else:
        ti = pow(t, i)
    if n == i and t == 1.0:
        tni = 1.0
    else:
        tni = pow((1.0 - t), (n - i))
    Ni = factorial(n) / (factorial(i) * factorial(n - i))
    return Ni * ti * tni


@lru_cache(maxsize=None)
def factorial(n: int):
    return math.factorial(n)



#  Copyright (c) 2021-2023, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.fonts.font_face import FontFace
from ezdxf.addons import text2path
from ezdxf.path import Path
from ezdxf import path, bbox
from ezdxf.entities import Text, Hatch
from ezdxf.layouts import VirtualLayout
from ezdxf.enums import TextEntityAlignment

# always available in common test scenarios:
DEFAULT = "LiberationSans-Regular.ttf"
NOTO_SANS_SC = "NotoSansSC-Regular.otf"


def _to_paths(s, f=DEFAULT):
    return text2path.make_paths_from_str(s, font=FontFace(filename=f))


# Font 'Arial' required, a replacement fonts may return a different
# path/hole configuration! issue #515
CHAR_TO_PATH = [
    ["1", 1],
    ["2", 1],
    [".", 1],
    ["0", 2],
    ["a", 2],
    ["!", 2],
    ["@", 2],
    ["8", 3],
    ["ü", 3],
    ["&", 3],
    ["ä", 4],
    ["ö", 4],
    ["%", 5],
]


@pytest.mark.parametrize("s,c", CHAR_TO_PATH)
def test_make_paths_from_str(s: str, c: int):
    # change of assertion from == to >= is based on an error on RPi & Ubuntu
    # Server 21.10 & matplotlib
    # assert len(_to_paths("ü")) == 4  instead of 3
    # Do not test how matplotlib works here, the only important fact is:
    # do we get path objects
    assert len(_to_paths(s)) >= c


@pytest.mark.parametrize("s,c", [["中", 3], ["国", 4], ["文", 3], ["字", 2]])
def test_chinese_char_paths_from_str(s, c):
    assert len(_to_paths(s, f=NOTO_SANS_SC)) == c


def contour_and_holes(group):
    return group[0], group[1:]


@pytest.mark.parametrize(
    "s,h",
    [
        ["1", 0],
        ["2", 0],
        [".", 0],
        ["0", 1],
        ["a", 1],
        ["8", 2],
    ],
)
def test_group_one_contour_with_holes(s, h):
    paths = _to_paths(s)
    result = list(path.group_paths(paths))
    contour, holes = contour_and_holes(result[0])
    assert isinstance(contour, Path)
    assert len(holes) == h


@pytest.mark.parametrize("s", [":", "!", ";", "="])
def test_group_two_contours_without_holes(s):
    paths = _to_paths(s)
    result = list(path.group_paths(paths))
    assert len(result) == 2
    contour, holes = contour_and_holes(result[0])
    assert isinstance(contour, Path)
    assert len(holes) == 0


@pytest.mark.parametrize(
    "s",
    [
        "Ü",
        "ö",
        "ä",
    ],
)
def test_group_three_contours_and_ignore_holes(s):
    paths = _to_paths(s)
    result = list(path.group_paths(paths))
    assert len(result) == 3
    contour, holes = contour_and_holes(result[0])
    assert isinstance(contour, Path)


@pytest.mark.parametrize("s,c", [["中", 1], ["国", 1], ["文", 2], ["字", 2]])
def test_group_chinese_chars_and_ignore_holes(s, c):
    paths = _to_paths(s, f=NOTO_SANS_SC)
    result = list(path.group_paths(paths))
    assert len(result) == c
    contour, holes = contour_and_holes(result[0])
    assert isinstance(contour, Path)


@pytest.fixture(scope="module")
def ff():
    return FontFace(family=DEFAULT)


class TestMakePathFromString:
    # Surprise - even 0 and negative values work without any exceptions!
    @pytest.mark.parametrize("size", [0, 0.05, 1, 2, 100, -1, -2, -100])
    def test_text_path_height_for_exact_drawing_units(self, size, ff):
        paths = text2path.make_paths_from_str("X", font=ff, size=size)
        bbox = path.bbox(paths)
        assert bbox.size.y == pytest.approx(abs(size))

    @pytest.mark.parametrize("size", [0.05, 1, 2, 100])
    def test_path_coordinates_for_positive_size(self, size, ff):
        paths = text2path.make_paths_from_str("X", font=ff, size=size)
        bbox = path.bbox(paths)
        assert bbox.extmax.y == pytest.approx(size)
        assert bbox.extmin.y == pytest.approx(0)

    @pytest.mark.parametrize("size", [-0.05, -1, -2, -100])
    def test_path_coordinates_for_negative_size(self, size, ff):
        # Negative text height mirrors text about the x-axis!
        paths = text2path.make_paths_from_str("X", font=ff, size=size)
        bbox = path.bbox(paths)
        assert bbox.extmax.y == pytest.approx(0)
        assert bbox.extmin.y == pytest.approx(size)

    @pytest.mark.parametrize("size", [0.05, 1, 2, 100])
    def test_length_for_fit_alignment(self, size, ff):
        length = 3
        paths = text2path.make_paths_from_str(
            "XXX",
            font=ff,
            size=size,
            align=TextEntityAlignment.FIT,
            length=length,
        )
        bbox = path.bbox(paths)
        assert bbox.size.x == pytest.approx(length), "expect exact length"
        assert bbox.size.y == pytest.approx(size), "text height should be unscaled"

    @pytest.mark.parametrize("size", [0.05, 1, 2, 100])
    def test_scaled_height_and_length_for_aligned_text(self, size, ff):
        length = 3
        paths = text2path.make_paths_from_str(
            "XXX", font=ff, size=size, align=TextEntityAlignment.LEFT
        )
        default = path.bbox(paths)
        paths = text2path.make_paths_from_str(
            "XXX",
            font=ff,
            size=size,
            align=TextEntityAlignment.ALIGNED,
            length=length,
        )
        bbox = path.bbox(paths)
        scale = bbox.size.x / default.size.x
        assert bbox.size.x == pytest.approx(length), "expect exact length"
        assert bbox.size.y == pytest.approx(
            size * scale
        ), "text height should be scaled"

    def test_paths_from_empty_string(self, ff):
        paths = text2path.make_paths_from_str("", font=ff)
        assert len(paths) == 0

    def test_make_multi_path_object(self, ff):
        p = text2path.make_path_from_str("ABC", font=ff)
        assert p.has_sub_paths is True
        assert len(list(p.sub_paths())) == 6

    def test_make_empty_multi_path_object(self, ff):
        p = text2path.make_path_from_str("", font=ff)
        assert p.has_sub_paths is False
        assert len(p) == 0


class TestMakeHatchesFromString:
    def test_hatches_from_empty_string(self, ff):
        hatches = text2path.make_hatches_from_str("", font=ff)
        assert len(hatches) == 0

    def test_make_exterior_only_hatches(self, ff):
        hatches = text2path.make_hatches_from_str("XXX", font=ff)
        assert len(hatches) == 3
        assert len(hatches[0].paths) == 1

    def test_make_hatches_with_holes(self, ff):
        hatches = text2path.make_hatches_from_str("AAA", font=ff)
        assert len(hatches) == 3
        assert len(hatches[0].paths) == 2, "expected external and one hole"

    def test_total_length_for_fit_alignment(self, ff):
        length = 3
        hatches = text2path.make_hatches_from_str(
            "XXX", font=ff, align=TextEntityAlignment.FIT, length=length
        )
        paths = []
        for hatch in hatches:
            paths.extend(path.from_hatch(hatch))
        bbox = path.bbox(paths)
        assert bbox.size.x == pytest.approx(length), "expect exact length"
        assert bbox.size.y == pytest.approx(1.0), "text height should be unscaled"


def test_check_entity_type():
    with pytest.raises(TypeError):
        text2path.check_entity_type(None)
    with pytest.raises(TypeError):
        text2path.check_entity_type(Hatch())


def make_text(text, location, alignment, height=1.0, rotation=0):
    text = Text.new(
        dxfattribs={
            "text": text,
            "height": height,
            "rotation": rotation,
        }
    )
    text.set_placement(location, align=alignment)
    return text


def get_path_bbox(text):
    p = text2path.make_path_from_entity(text)
    return path.bbox([p], fast=True)


def get_paths_bbox(text):
    paths = text2path.make_paths_from_entity(text)
    return path.bbox(paths, fast=True)


def get_hatches_bbox(text):
    hatches = text2path.make_hatches_from_entity(text)
    return bbox.extents(hatches, fast=True)


@pytest.fixture(params=[get_path_bbox, get_paths_bbox, get_hatches_bbox])
def get_bbox(request):
    return request.param


class TestMakePathsFromEntity:
    """Test Paths (and Hatches) from TEXT entities.

    make_hatches_from_entity() is basically make_paths_from_entity(), but
    returns Hatch entities instead of Path objects.

    Important: Don't use text with top or bottom curves for testing ("S", "O").
    The Path bounding box calculation uses the "fast" method by checking only
    the curve control points, which are outside the curve borders.

    """

    @pytest.mark.parametrize(
        "builder, type_",
        [
            (text2path.make_paths_from_entity, Path),
            (text2path.make_hatches_from_entity, Hatch),
        ],
    )
    def test_text_returns_correct_types(self, builder, type_):
        text = make_text("TEXT", (0, 0), TextEntityAlignment.LEFT)
        objects = builder(text)
        assert len(objects) == 4
        assert isinstance(objects[0], type_)

    def test_text_height(self, get_bbox):
        text = make_text("TEXT", (0, 0), TextEntityAlignment.LEFT, height=1.5)
        bbox = get_bbox(text)
        assert bbox.size.y == pytest.approx(1.5)

    def test_alignment_left(self, get_bbox):
        text = make_text("TEXT", (7, 7), TextEntityAlignment.LEFT)
        bbox = get_bbox(text)
        # font rendering is tricky, base offsets depend on the rendering engine
        # and on extended font metrics, ...
        assert bbox.extmin.x == pytest.approx(7, abs=0.1)

    def test_alignment_center(self, get_bbox):
        text = make_text("TEXT", (7, 7), TextEntityAlignment.CENTER)
        bbox = get_bbox(text)
        assert bbox.center.x == pytest.approx(7)

    def test_alignment_right(self, get_bbox):
        text = make_text("TEXT", (7, 7), TextEntityAlignment.RIGHT)
        bbox = get_bbox(text)
        assert bbox.extmax.x == pytest.approx(7)

    def test_alignment_baseline(self, get_bbox):
        text = make_text("TEXT", (7, 7), TextEntityAlignment.CENTER)
        bbox = get_bbox(text)
        assert bbox.extmin.y == pytest.approx(7)

    def test_alignment_bottom(self, get_bbox):
        text = make_text("j", (7, 7), TextEntityAlignment.BOTTOM_CENTER)
        bbox = get_bbox(text)
        # bottom border of descender should be 7, but ...
        assert bbox.extmin.y == pytest.approx(7, abs=0.1)

    def test_alignment_middle(self, get_bbox):
        text = make_text("X", (7, 7), TextEntityAlignment.MIDDLE_CENTER)
        bbox = get_bbox(text)
        assert bbox.center.y == pytest.approx(7)

    def test_alignment_top(self, get_bbox):
        text = make_text("X", (7, 7), TextEntityAlignment.TOP_CENTER)
        bbox = get_bbox(text)
        assert bbox.extmax.y == pytest.approx(7)

    def test_alignment_fit(self, get_bbox):
        length = 2
        height = 1
        text = make_text("TEXT", (0, 0), TextEntityAlignment.LEFT, height=height)
        text.set_placement((1, 0), (1 + length, 0), TextEntityAlignment.FIT)
        bbox = get_bbox(text)
        assert bbox.size.x == length, "expected text length fits into given length"
        assert bbox.size.y == pytest.approx(height), "expected unscaled text height"
        assert bbox.extmin.isclose((1, 0))

    def test_alignment_aligned(self, get_bbox):
        length = 2
        height = 1
        text = make_text("TEXT", (0, 0), TextEntityAlignment.CENTER, height=height)
        bbox = get_bbox(text)
        ratio = bbox.size.x / bbox.size.y

        text.set_placement((1, 0), (1 + length, 0), TextEntityAlignment.ALIGNED)
        bbox = get_bbox(text)

        assert bbox.size.x == length, "expected text length fits into given length"
        assert bbox.size.y != height, "expected scaled text height"
        assert bbox.extmin.isclose((1, 0))
        assert bbox.size.x / bbox.size.y == pytest.approx(
            ratio
        ), "expected same width/height ratio"

    def test_rotation_90(self, get_bbox):
        # Horizontal reference measurements:
        bbox_hor = get_bbox(
            make_text("TEXT", (7, 7), TextEntityAlignment.MIDDLE_CENTER)
        )

        text_vert = make_text(
            "TEXT", (7, 7), TextEntityAlignment.MIDDLE_CENTER, rotation=90
        )
        bbox_vert = get_bbox(text_vert)
        assert bbox_hor.center == bbox_vert.center
        assert bbox_hor.size.x == bbox_vert.size.y
        assert bbox_hor.size.y == bbox_vert.size.x


Kind = text2path.Kind


class TestVirtualEntities:
    @pytest.fixture
    def text(self):
        return make_text("TEST", (0, 0), TextEntityAlignment.LEFT)

    def test_virtual_entities_as_hatches(self, text):
        entities = text2path.virtual_entities(text, kind=Kind.HATCHES)
        types = {e.dxftype() for e in entities}
        assert types == {"HATCH"}

    def test_virtual_entities_as_splines_and_polylines(self, text):
        entities = text2path.virtual_entities(text, kind=Kind.SPLINES)
        types = {e.dxftype() for e in entities}
        assert types == {"SPLINE", "POLYLINE"}

    def test_virtual_entities_as_lwpolylines(self, text):
        entities = text2path.virtual_entities(text, kind=Kind.LWPOLYLINES)
        types = {e.dxftype() for e in entities}
        assert types == {"LWPOLYLINE"}

    def test_virtual_entities_to_all_types_at_once(self, text):
        entities = text2path.virtual_entities(
            text, kind=Kind.HATCHES + Kind.SPLINES + Kind.LWPOLYLINES
        )
        types = {e.dxftype() for e in entities}
        assert types == {"LWPOLYLINE", "SPLINE", "POLYLINE", "HATCH"}


class TestExplode:
    """Based on text2path.virtual_entities() function, see test above."""

    @pytest.fixture
    def text(self):
        return make_text("TEST", (0, 0), TextEntityAlignment.LEFT)

    def test_source_entity_is_destroyed(self, text):
        assert text.is_alive is True
        text2path.explode(text, kind=4)
        assert text.is_alive is False, "source entity should always be destroyed"

    def test_explode_entity_into_layout(self, text):
        layout = VirtualLayout()
        entities = text2path.explode(text, kind=Kind.LWPOLYLINES, target=layout)
        assert len(entities) == len(
            layout
        ), "expected all entities added to the target layout"

    def test_explode_entity_into_the_void(self, text):
        assert text.get_layout() is None, "source entity should not have a layout"
        entities = text2path.explode(text, kind=Kind.LWPOLYLINES, target=None)
        assert len(entities) == 4, "explode should work without a target layout"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (C) 2011-2021, Manfred Moitzi
# License: MIT License
import pytest
from io import StringIO

import ezdxf
from ezdxf.entities.dimstyle import DimStyle
from ezdxf.document import Drawing
from ezdxf.lldxf.const import DXF12
from ezdxf.lldxf.tagwriter import TagCollector, TagWriter

TEXT_STYLE = "TextStyle"


@pytest.fixture
def dimstyle():
    doc = Drawing.new()
    setup_doc(doc)
    return DimStyle.new(
        "FFFF",
        doc=doc,
        dxfattribs={
            "name": "DIMSTYLE1",
        },
    )


def setup_doc(doc):
    doc.blocks.new("left_arrow")
    doc.blocks.new("right_arrow")
    doc.blocks.new("arrow")
    doc.blocks.new("_DOT")
    doc.blocks.new("_OPEN")
    doc.styles.new(TEXT_STYLE)


def test_name(dimstyle):
    assert "DIMSTYLE1" == dimstyle.dxf.name


def test_set_blk1_and_blk2_arrows(dimstyle):
    dimstyle.set_arrows("", "left_arrow", "right_arrow")
    assert dimstyle.dxf.dimblk == ""
    assert dimstyle.dxf.dimblk1 == "left_arrow"
    assert dimstyle.dxf.dimblk2 == "right_arrow"


def test_set_both_arrows(dimstyle):
    dimstyle.set_arrows("arrow")
    assert dimstyle.dxf.dimblk == "arrow"
    assert dimstyle.dxf.dimblk1 == ""
    assert dimstyle.dxf.dimblk2 == ""

    dimstyle.set_arrows(blk1="OPEN", blk2="DOT")
    assert dimstyle.dxf.dimblk == ""
    assert dimstyle.dxf.dimblk1 == "OPEN"
    assert dimstyle.dxf.dimblk2 == "DOT"


def test_set_tick(dimstyle):
    dimstyle.set_tick(0.25)
    assert dimstyle.dxf.dimtsz == 0.25


def test_set_text_align(dimstyle):
    dimstyle.set_text_align(valign="above")
    assert dimstyle.dxf.dimtad == 1


def test_set_text_format(dimstyle):
    dimstyle.set_text_format(
        prefix="+",
        postfix=" cm",
        rnd=0.5,
        leading_zeros=False,
        trailing_zeros=False,
    )
    assert dimstyle.dxf.dimpost == "+<> cm"
    assert dimstyle.dxf.dimrnd == 0.5
    assert dimstyle.dxf.dimzin == 12


@pytest.fixture(scope="module")
def dimstyle2():
    import ezdxf

    doc = ezdxf.new("R2007", setup=("linetypes",))
    doc.blocks.new("left_arrow")
    doc.blocks.new("right_arrow")
    doc.blocks.new("TestArrow")
    doc.blocks.new("arrow")
    doc.styles.new(TEXT_STYLE)
    return doc.dimstyles.new("testing")


def test_handle_export(dimstyle2):
    textstyle = dimstyle2.doc.styles.get(TEXT_STYLE)
    dimstyle2.set_arrows("", "left_arrow", "right_arrow", "arrow")
    dimstyle2.dxf.dimtxsty = textstyle.dxf.name

    # test handles
    blocks = dimstyle2.doc.blocks
    left_arrow = blocks.get("left_arrow")
    right_arrow = blocks.get("right_arrow")
    leader_arrow = blocks.get("arrow")

    # prepare for export
    dimstyle2.set_handles()
    assert dimstyle2.dxf.hasattr("dimblk_handle") is False
    assert dimstyle2.dxf.dimblk1_handle == left_arrow.block_record_handle
    assert dimstyle2.dxf.dimblk2_handle == right_arrow.block_record_handle
    assert dimstyle2.dxf.dimldrblk_handle == leader_arrow.block_record_handle
    assert dimstyle2.dxf.dimtxsty_handle == textstyle.dxf.handle


def test_resource_handles_export_R2000():
    doc = ezdxf.new()
    setup_doc(doc)
    dimstyle = doc.dimstyles.add("MyStyle")
    dimstyle.set_arrows("arrow", "left_arrow", "right_arrow", "arrow")
    dimstyle.dxf.dimtxsty = TEXT_STYLE
    t = TagCollector()
    dimstyle.export_dxf(t)
    codes = {
        340,  # dimtxsty_handle
        341,  # dimldrblk_handle
        342,  # dimblk_handle
        343,  # dimblk1_handle
        344,  # dimblk2_handle
    }
    handles = [tag.value for tag in t.tags if tag.code in codes]
    assert len(handles) == 5, "5 handles expected"
    for handle in handles:
        assert handle in doc.entitydb, f"not a valid handle: {handle}"


def test_text_stream_for_resource_handles_R2000():
    s = StringIO()
    t = TagWriter(s)

    doc = ezdxf.new()
    setup_doc(doc)
    dimstyle = doc.dimstyles.add("MyStyle")
    dimstyle.set_arrows("arrow", "left_arrow", "right_arrow", "arrow")
    dimstyle.dxf.dimtxsty = TEXT_STYLE

    dimstyle.export_dxf(t)
    result = s.getvalue()
    assert "\n340\n" in result
    assert "\n341\n" in result
    assert "\n342\n" in result
    assert "\n343\n" in result
    assert "\n344\n" in result


def test_dont_write_handles_for_R12(dimstyle):
    s = StringIO()
    t = TagWriter(s)
    t.dxfversion = DXF12
    t.write_handles = False
    dimstyle.export_dxf(t)
    result = s.getvalue()
    assert "105\nFFFF\n" not in result


def test_dimstyle_name(dimstyle2):
    assert "testing" == dimstyle2.dxf.name


def test_dimstyle_blk1_and_blk2_ticks(dimstyle2):
    dimstyle2.set_arrows("", "left_arrow", "right_arrow")
    assert dimstyle2.get_dxf_attrib("dimblk") == ""
    assert dimstyle2.get_dxf_attrib("dimblk1") == "left_arrow"
    assert dimstyle2.get_dxf_attrib("dimblk2") == "right_arrow"


def test_dimstyle_both_ticks(dimstyle2):
    dimstyle2.set_arrows("arrow")
    assert dimstyle2.get_dxf_attrib("dimblk") == "arrow"
    assert dimstyle2.get_dxf_attrib("dimblk1") == ""  # closed filled
    assert dimstyle2.get_dxf_attrib("dimblk2") == ""  # closed filled

    assert dimstyle2.get_dxf_attrib("dimblk1") == ""
    assert dimstyle2.get_dxf_attrib("dimblk2") == ""


def test_dimstyle_virtual_dimtxsty_attribute(dimstyle2):
    dimstyle2.dxf.dimtxsty = TEXT_STYLE
    assert dimstyle2.dxf.dimtxsty == TEXT_STYLE

    # prepare for export
    dimstyle2.set_handles()
    assert (
        dimstyle2.dxf.dimtxsty_handle
        == dimstyle2.doc.styles.get(TEXT_STYLE).dxf.handle
    )


def test_dimstyle_virtual_dimldrblk_attribute(dimstyle2):
    dimstyle2.dxf.dimldrblk = "CLOSED"

    dimstyle2.dxf.dimldrblk = "TestArrow"
    assert dimstyle2.dxf.dimldrblk == "TestArrow"


def test_dimstyle_virtual_linetypes_attributes(dimstyle2):
    dimstyle2.dxf.dimltype = "DOT2"
    assert dimstyle2.dxf.dimltype == "DOT2"

    dimstyle2.dxf.dimltex1 = "DOT"
    assert dimstyle2.dxf.dimltex1 == "DOT"

    dimstyle2.dxf.dimltex2 = "DOTX2"
    assert dimstyle2.dxf.dimltex2 == "DOTX2"

    # prepare for export
    dimstyle2.set_handles()
    assert (
        dimstyle2.dxf.dimltype_handle
        == dimstyle2.doc.linetypes.get("DOT2").dxf.handle
    )
    assert (
        dimstyle2.dxf.dimltex1_handle
        == dimstyle2.doc.linetypes.get("DOT").dxf.handle
    )
    assert (
        dimstyle2.dxf.dimltex2_handle
        == dimstyle2.doc.linetypes.get("DOTX2").dxf.handle
    )


def test_dimstyle_group_codes(dimstyle):
    codes = DimStyle.CODE_TO_DXF_ATTRIB
    assert 105 not in codes
    assert 2 not in codes


def test_dimstyle_set_align(dimstyle2):
    dimstyle2.set_text_align(valign="above")
    assert dimstyle2.dxf.dimtad == 1
    dimstyle2.set_text_align(halign="above1")
    assert dimstyle2.dxf.dimjust == 3


def test_set_text_format_2(dimstyle2):
    dimstyle2.set_text_format(
        prefix="+",
        postfix=" cm",
        rnd=0.5,
        dec=2,
        sep=".",
        leading_zeros=False,
        trailing_zeros=False,
    )
    assert dimstyle2.dxf.dimpost == "+<> cm"
    assert dimstyle2.dxf.dimrnd == 0.5
    assert dimstyle2.dxf.dimdec == 2
    assert dimstyle2.dxf.dimdsep == ord(".")
    assert dimstyle2.dxf.dimzin == 12


def test_set_dimline_format(dimstyle2):
    dimstyle2.set_dimline_format(
        color=2,
        linetype="DOT",
        lineweight=18,
        extension=0.33,
        disable1=True,
        disable2=True,
    )
    assert dimstyle2.dxf.dimclrd == 2
    assert dimstyle2.dxf.dimltype == "DOT"
    assert dimstyle2.dxf.dimlwd == 18
    assert dimstyle2.dxf.dimdle == 0.33
    assert dimstyle2.dxf.dimsd1 == 1
    assert dimstyle2.dxf.dimsd2 == 1


def test_set_extline_format(dimstyle2):
    dimstyle2.set_extline_format(
        color=2,
        lineweight=18,
        extension=0.33,
        offset=0.77,
        fixed_length=0.5,
    )
    assert dimstyle2.dxf.dimclre == 2
    assert dimstyle2.dxf.dimlwe == 18
    assert dimstyle2.dxf.dimexe == 0.33
    assert dimstyle2.dxf.dimexo == 0.77
    assert dimstyle2.dxf.dimfxlon == 1
    assert dimstyle2.dxf.dimfxl == 0.5


def test_set_extline1(dimstyle2):
    dimstyle2.set_extline1(
        linetype="DOT",
        disable=True,
    )
    assert dimstyle2.dxf.dimltex1 == "DOT"
    assert dimstyle2.dxf.dimse1 == 1


def test_set_extline2(dimstyle2):
    dimstyle2.set_extline2(
        linetype="DOT",
        disable=True,
    )
    assert dimstyle2.dxf.dimltex2 == "DOT"
    assert dimstyle2.dxf.dimse2 == 1



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
# created 2019-02-14
import pytest
import ezdxf

from ezdxf.lldxf.const import DXFAttributeError, DXF12, DXFValueError
from ezdxf.lldxf.tagwriter import TagCollector
from ezdxf.entities import DXFEntity, is_graphic_entity, Insert
from ezdxf.lldxf.extendedtags import DXFTag
from ezdxf.entities.line import Line

ENTITY = """0
DXFENTITY
5
FFFF
330
ABBA
"""


@pytest.fixture
def entity():
    return DXFEntity.from_text(ENTITY)


def test_dxfentity_is_set_compatible():
    assert len({DXFEntity(), DXFEntity(), DXFEntity()}) == 3


def test_is_not_graphic_entity(entity):
    assert is_graphic_entity(entity) is False


def test_default_constructor():
    entity = DXFEntity()
    assert entity.dxftype() == "DXFENTITY"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None
    assert entity == entity
    assert entity != DXFEntity()


def test_init_with_tags(entity):
    assert entity.dxftype() == "DXFENTITY"
    assert entity.dxf.handle == "FFFF"
    assert entity.dxf.owner == "ABBA"
    assert str(entity) == "DXFENTITY(#FFFF)"
    assert (
        repr(entity)
        == "<class 'ezdxf.entities.dxfentity.DXFEntity'> DXFENTITY(#FFFF)"
    )


def test_invalid_dxf_attrib(entity):
    with pytest.raises(DXFAttributeError):
        _ = entity.dxf.color


def test_get_all_dxf_attribs(entity):
    dxfattribs = entity.dxfattribs()
    assert len(dxfattribs) == 2
    assert dxfattribs["handle"] == "FFFF"
    assert dxfattribs["owner"] == "ABBA"


def test_get_dxf_attribs_and_discard_some(entity):
    dxfattribs = entity.dxfattribs(drop={"owner"})
    assert len(dxfattribs) == 1
    assert dxfattribs["handle"] == "FFFF"
    assert "owner" not in dxfattribs


def test_write_r12_dxf(entity):
    tagwriter = TagCollector(dxfversion=DXF12)
    entity.export_dxf(tagwriter)
    tag = tagwriter.tags
    assert len(tag) == 2
    assert tag[0] == (0, "DXFENTITY")
    assert tag[1] == (5, "FFFF")


def test_write_latest_dxf(entity):
    tagwriter = TagCollector()
    entity.export_dxf(tagwriter)
    tag = tagwriter.tags
    assert len(tag) == 3
    assert tag[0] == (0, "DXFENTITY")
    assert tag[1] == (5, "FFFF")
    assert tag[2] == (330, "ABBA")


def test_is_alive(entity):
    assert entity.is_alive is True
    entity.destroy()
    assert entity.is_alive is False


def test_calling_destroy_multiple_times(entity):
    entity.destroy()
    entity.destroy(), "2nd call should do nothing"
    entity.destroy(), "3rd call should do nothing"
    assert entity.is_alive is False


def test_dont_write_handles_for_R12(entity):
    from ezdxf.lldxf.tagwriter import TagWriter
    from io import StringIO

    s = StringIO()
    t = TagWriter(s)
    t.dxfversion = DXF12
    t.write_handles = False
    entity.export_dxf(t)
    result = s.getvalue()
    assert "5\nFFFF\n" not in result


def test_uuid():
    e1 = DXFEntity()
    e2 = DXFEntity()
    assert e1.dxf.handle == e2.dxf.handle, "can't distinguish by handle"
    assert e1.uuid != e2.uuid, "can distinguish by uuid"


def test_source_of_copy_is_none_for_a_new_entity():
    e = DXFEntity()
    assert e.source_of_copy is None


def test_set_source_of_copy():
    e = DXFEntity()
    e.set_source_of_copy(e)
    assert e.source_of_copy is e


def test_delete_missing_source_of_copy_without_exception():
    e = DXFEntity()
    e.del_source_of_copy()
    assert True is True


def test_source_block_reference_is_none_for_a_new_entity():
    e = DXFEntity()
    assert e.has_source_block_reference is False
    assert e.source_block_reference is None


def test_set_source_block_reference():
    e = DXFEntity()
    insert = Insert()
    e.set_source_block_reference(insert)
    assert e.has_source_block_reference is True
    assert e.source_block_reference is insert


def test_setting_source_block_reference_twice_without_exception():
    e = DXFEntity()
    e.set_source_block_reference(Insert())
    e.set_source_block_reference(Insert())
    assert True is True


def test_setting_source_block_reference_a_second_time_has_no_effect():
    e = DXFEntity()
    insert = Insert()
    e.set_source_block_reference(insert)
    e.set_source_block_reference(Insert())
    assert (
        e.source_block_reference is insert
    ), "source block reference should not change"


def test_do_not_copy_source_block_reference():
    e = DXFEntity()
    insert = Insert()
    e.set_source_block_reference(insert)
    copy = e.copy()
    assert copy.has_source_block_reference is False
    assert copy.source_block_reference is None


def test_delete_missing_source_block_reference_without_exception():
    e = DXFEntity()
    e.del_source_block_reference()
    assert True is True


LINE_DATA = """  0
LINE
  5
0
330
0
100
AcDbEntity
  8
0
100
AcDbLine
 10
0.0
 20
0.0
 30
0.0
 11
1.0
 21
1.0
 31
1.0
"""


@pytest.fixture
def line():
    return Line.from_text(LINE_DATA)


def test_str(line):
    assert str(line) == "LINE(#0)"
    assert repr(line) == "<class 'ezdxf.entities.line.Line'> LINE(#0)"


def test_get_dxf_defaul(line):
    # get_dxf_default returns the DXF default value for unset attributes
    assert line.dxf.get_default("thickness") == 0
    # get returns the given default for unset attributes, which is None by default :)
    assert line.dxf.get("thickness") is None


def test_ocs(line):
    ocs = line.ocs()
    assert ocs.uz == (0, 0, 1)
    assert ocs.transform is False


class TestAppData:
    @pytest.fixture
    def entity(self):
        return Line.from_text(LINE_DATA)

    def test_new_app_data(self, entity):
        assert entity.has_app_data("{MOZMAN") is False
        entity.set_app_data("{MOZMAN", tags=[DXFTag(330, "DEAD")])
        assert entity.has_app_data("{MOZMAN") is True

    def test_get_app_data(self, entity):
        entity.set_app_data("{MOZMAN", tags=[DXFTag(330, "DEAD")])

        app_data = entity.get_app_data("{MOZMAN")
        assert len(app_data) == 1
        assert DXFTag(330, "DEAD") == app_data[0]

    def test_set_app_data(self, entity):
        entity.set_app_data("{MOZMAN", tags=[DXFTag(330, "DEAD")])
        app_data = entity.get_app_data("{MOZMAN")
        assert 1 == len(app_data)
        assert DXFTag(330, "DEAD") == app_data[0]
        app_data.append(DXFTag(360, "DEAD2"))
        entity.set_app_data("{MOZMAN", app_data)

        app_data = entity.get_app_data("{MOZMAN")
        assert 2 == len(app_data)
        assert DXFTag(330, "DEAD") == app_data[0]
        assert DXFTag(360, "DEAD2") == app_data[1]

    def test_not_existing_appid(self, entity):
        with pytest.raises(DXFValueError):
            entity.get_app_data("XYZ")


class TestXData:
    @pytest.fixture
    def entity(self):
        return Line.from_text(LINE_DATA)

    def test_new_app_data(self, entity):
        assert entity.has_xdata("MOZMAN") is False
        entity.set_xdata("MOZMAN", tags=[DXFTag(1000, "Extended Data String")])
        assert entity.has_xdata("MOZMAN") is True

    def test_get_xdata(self, entity):
        entity.set_xdata("MOZMAN", tags=[DXFTag(1000, "Extended Data String")])

        xdata = entity.get_xdata("MOZMAN")
        assert len(xdata) == 1
        assert DXFTag(1000, "Extended Data String") == xdata[0]

    def test_set_xdata(self, entity):
        entity.set_xdata("MOZMAN", tags=[DXFTag(1000, "Extended Data String")])
        xdata = entity.get_xdata("MOZMAN")
        assert 1 == len(xdata)
        assert DXFTag(1000, "Extended Data String") == xdata[0]
        xdata.append(DXFTag(1000, "Extended Data String2"))
        entity.set_xdata("MOZMAN", xdata)

        xdata = entity.get_xdata("MOZMAN")
        assert 2 == len(xdata)
        assert DXFTag(1000, "Extended Data String") == xdata[0]
        assert DXFTag(1000, "Extended Data String2") == xdata[1]

    def test_not_existing_appid(self, entity):
        with pytest.raises(DXFValueError):
            entity.get_xdata("XYZ")

    def test_get_xdata_list_exception(self, entity):
        with pytest.raises(DXFValueError):
            _ = entity.get_xdata_list("ACAD", "DSTYLE")
        entity.set_xdata("ACAD", tags=[DXFTag(1000, "Extended Data String")])

        with pytest.raises(DXFValueError):
            _ = entity.get_xdata_list("ACAD", "DSTYLE")

    def test_has_xdata_list(self, entity):
        assert entity.has_xdata_list("ACAD", "DSTYLE") is False
        entity.set_xdata("ACAD", tags=[DXFTag(1000, "Extended Data String")])
        assert entity.has_xdata_list("ACAD", "DSTYLE") is False

    def test_set_xdata_list(self, entity):
        entity.set_xdata_list("ACAD", "DSTYLE", [(1070, 1), (1000, "String")])
        xdata_list = entity.get_xdata_list("ACAD", "DSTYLE")
        assert len(xdata_list) == 5
        assert xdata_list == [
            (1000, "DSTYLE"),
            (1002, "{"),
            (1070, 1),
            (1000, "String"),
            (1002, "}"),
        ]
        # add another list to ACAD
        entity.set_xdata_list("ACAD", "MOZMAN", [(1070, 2), (1000, "mozman")])
        xdata = entity.get_xdata_list("ACAD", "MOZMAN")
        assert len(xdata) == 5
        assert xdata == [
            (1000, "MOZMAN"),
            (1002, "{"),
            (1070, 2),
            (1000, "mozman"),
            (1002, "}"),
        ]
        xdata = entity.get_xdata("ACAD")
        assert len(xdata) == 10

    def test_discard_xdata_list(self, entity):
        entity.set_xdata_list("ACAD", "DSTYLE", [(1070, 1), (1000, "String")])
        xdata_list = entity.get_xdata_list("ACAD", "DSTYLE")
        assert len(xdata_list) == 5
        entity.discard_xdata_list("ACAD", "DSTYLE")
        with pytest.raises(DXFValueError):
            _ = entity.get_xdata_list("ACAD", "DSTYLE")

        entity.discard_xdata_list("ACAD", "DSTYLE")

    def test_replace_xdata_list(self, entity):
        entity.set_xdata_list("ACAD", "DSTYLE", [(1070, 1), (1000, "String")])
        xdata_list = entity.get_xdata_list("ACAD", "DSTYLE")
        assert len(xdata_list) == 5
        assert xdata_list == [
            (1000, "DSTYLE"),
            (1002, "{"),
            (1070, 1),
            (1000, "String"),
            (1002, "}"),
        ]
        entity.set_xdata_list(
            "ACAD", "DSTYLE", [(1070, 2), (1000, "mozman"), (1000, "data")]
        )
        xdata_list = entity.get_xdata_list("ACAD", "DSTYLE")
        assert len(xdata_list) == 6
        assert xdata_list == [
            (1000, "DSTYLE"),
            (1002, "{"),
            (1070, 2),
            (1000, "mozman"),
            (1000, "data"),
            (1002, "}"),
        ]
        # replace not existing list -> append list
        entity.replace_xdata_list("ACAD", "MOZMAN", [(1070, 3), (1000, "new")])
        xdata_list = entity.get_xdata_list("ACAD", "MOZMAN")
        assert len(xdata_list) == 5
        assert xdata_list == [
            (1000, "MOZMAN"),
            (1002, "{"),
            (1070, 3),
            (1000, "new"),
            (1002, "}"),
        ]
        xdata = entity.get_xdata("ACAD")
        assert len(xdata) == 6 + 5


class TestReactors:
    @pytest.fixture
    def entity(self):
        return Line.from_text(LINE_DATA)

    def test_has_no_reactors(self, entity):
        assert entity.has_reactors() is False

    def test_set_reactors(self, entity):
        entity.set_reactors(["A000", "B000", "C000"])
        assert entity.has_reactors() is True
        handles = entity.get_reactors()
        assert ["A000", "B000", "C000"] == handles

    def test_append_handle(self, entity):
        entity.set_reactors([])
        assert 0 == len(entity.get_reactors())
        entity.append_reactor_handle("A000")
        assert "A000" in entity.get_reactors()
        entity.append_reactor_handle("B000")
        assert "B000" in entity.get_reactors()
        assert 2 == len(entity.get_reactors())

        entity.append_reactor_handle("B000")  # same handle again
        assert "B000" in entity.get_reactors()
        assert 2 == len(
            entity.get_reactors()
        ), "handle entries should be unique"

        entity.append_reactor_handle(
            "FF"
        )  # smallest handle, should be first handle in reactors
        assert entity.get_reactors()[0] == "FF"

        entity.append_reactor_handle(
            "FFFF"
        )  # biggest handle, should be last handle in reactors
        assert "FFFF" == entity.get_reactors()[-1]

    def test_discard_handle(self, entity):
        entity.set_reactors(["A000", "B000", "C000"])
        entity.discard_reactor_handle("A000")
        assert 2 == len(entity.get_reactors()), "Handle not deleted"
        entity.discard_reactor_handle("FFFF")  # ignore not existing handles
        assert 2 == len(entity.get_reactors())


class TestGetLayout:
    @pytest.fixture(scope="class")
    def doc(self):
        return ezdxf.new()

    def test_get_layout_model_space(self, doc):
        msp = doc.modelspace()
        circle = msp.add_circle(center=(0, 0), radius=1)
        layout = circle.get_layout()
        assert msp.name == layout.name

    def test_get_layout_paper_space(self, doc):
        psp = doc.layout()
        circle = psp.add_circle(center=(0, 0), radius=1)
        layout = circle.get_layout()
        assert psp.name == layout.name

    def test_get_layout_block(self, doc):
        block = doc.blocks.new("TEST")
        circle = block.add_circle(center=(0, 0), radius=1)
        layout = circle.get_layout()
        assert block.name == layout.name

    def test_get_layout_without_layout(self, doc):
        msp = doc.modelspace()
        circle = msp.add_circle(center=(0, 0), radius=1)
        msp.unlink_entity(circle)
        assert circle.get_layout() is None



# Copyright (c) 2018-2020, Manfred Moitzi
# License: MIT License

import pytest
import ezdxf
import math
from ezdxf.math import (
    ConstructionArc,
    ConstructionCircle,
    ConstructionLine,
    UCS,
    Vec3,
    Vec2,
    arc_segment_count,
    arc_chord_length,
)
from math import isclose


def test_arc_from_2p_angle_complex():
    p1 = (-15.73335, 10.98719)
    p2 = (-12.67722, 8.76554)
    angle = 55.247230
    arc = ConstructionArc.from_2p_angle(start_point=p1, end_point=p2, angle=angle)

    arc_result = ConstructionArc(
        center=(-12.08260, 12.79635),
        radius=4.07443,
        start_angle=-153.638906,
        end_angle=-98.391676,
    )

    assert arc.center.isclose(arc_result.center, abs_tol=1e-5)
    assert isclose(arc.radius, arc_result.radius, abs_tol=1e-5)
    assert isclose(arc.start_angle, arc_result.start_angle, abs_tol=1e-4)
    assert isclose(arc.end_angle, arc_result.end_angle, abs_tol=1e-4)


def test_arc_from_2p_angle_simple():
    p1 = (2, 1)
    p2 = (0, 3)
    angle = 90

    arc = ConstructionArc.from_2p_angle(start_point=p1, end_point=p2, angle=angle)
    assert arc.center.isclose((0, 1))
    assert isclose(arc.radius, 2)
    assert isclose(arc.start_angle, 0, abs_tol=1e-12)
    assert isclose(arc.end_angle, 90)

    arc = ConstructionArc.from_2p_angle(start_point=p2, end_point=p1, angle=angle)
    assert arc.center.isclose((2, 3))
    assert isclose(arc.radius, 2)
    assert isclose(arc.start_angle, 180)
    assert isclose(arc.end_angle, -90)


def test_arc_from_2p_radius():
    p1 = (2, 1)
    p2 = (0, 3)
    radius = 2

    arc = ConstructionArc.from_2p_radius(start_point=p1, end_point=p2, radius=radius)
    assert arc.center.isclose((0, 1))
    assert isclose(arc.radius, radius)
    assert isclose(arc.start_angle, 0)
    assert isclose(arc.end_angle, 90)

    arc = ConstructionArc.from_2p_radius(start_point=p2, end_point=p1, radius=radius)
    assert arc.center.isclose((2, 3))
    assert isclose(arc.radius, radius)
    assert isclose(arc.start_angle, 180)
    assert isclose(arc.end_angle, -90)


def test_arc_from_3p():
    p1 = (-15.73335, 10.98719)
    p2 = (-12.67722, 8.76554)
    p3 = (-8.00817, 12.79635)
    arc = ConstructionArc.from_3p(start_point=p1, end_point=p2, def_point=p3)

    arc_result = ConstructionArc(
        center=(-12.08260, 12.79635),
        radius=4.07443,
        start_angle=-153.638906,
        end_angle=-98.391676,
    )

    assert arc.center.isclose(arc_result.center, abs_tol=1e-5)
    assert isclose(arc.radius, arc_result.radius, abs_tol=1e-5)
    assert isclose(arc.start_angle, arc_result.start_angle, abs_tol=1e-4)
    assert isclose(arc.end_angle, arc_result.end_angle, abs_tol=1e-4)


def test_spatial_arc_from_3p():
    start_point_wcs = Vec3(0, 1, 0)
    end_point_wcs = Vec3(1, 0, 0)
    def_point_wcs = Vec3(0, 0, 1)

    ucs = UCS.from_x_axis_and_point_in_xy(
        origin=def_point_wcs,
        axis=end_point_wcs - def_point_wcs,
        point=start_point_wcs,
    )
    start_point_ucs = ucs.from_wcs(start_point_wcs)
    end_point_ucs = ucs.from_wcs(end_point_wcs)
    def_point_ucs = Vec3(0, 0)

    arc = ConstructionArc.from_3p(start_point_ucs, end_point_ucs, def_point_ucs)
    dwg = ezdxf.new("R12")
    msp = dwg.modelspace()

    dxf_arc = arc.add_to_layout(msp, ucs)
    assert dxf_arc.dxftype() == "ARC"
    assert isclose(dxf_arc.dxf.radius, 0.81649658, abs_tol=1e-9)
    assert isclose(dxf_arc.dxf.start_angle, -30)
    assert isclose(dxf_arc.dxf.end_angle, -150)
    assert dxf_arc.dxf.extrusion.isclose(
        (0.57735027, 0.57735027, 0.57735027), abs_tol=1e-9
    )


def test_bounding_box():
    bbox = ConstructionArc(
        center=(0, 0), radius=1, start_angle=0, end_angle=90
    ).bounding_box
    assert bbox.extmin.isclose((0, 0))
    assert bbox.extmax.isclose((1, 1))

    bbox = ConstructionArc(
        center=(0, 0), radius=1, start_angle=0, end_angle=180
    ).bounding_box
    assert bbox.extmin.isclose((-1, 0))
    assert bbox.extmax.isclose((1, 1))

    bbox = ConstructionArc(
        center=(0, 0), radius=1, start_angle=270, end_angle=90
    ).bounding_box
    assert bbox.extmin.isclose((0, -1))
    assert bbox.extmax.isclose((1, 1))


def test_angles():
    arc = ConstructionArc(radius=1, start_angle=30, end_angle=60)
    assert tuple(arc.angles(2)) == (30, 60)
    assert tuple(arc.angles(3)) == (30, 45, 60)

    arc.start_angle = 180
    arc.end_angle = 0
    assert tuple(arc.angles(2)) == (180, 0)
    assert tuple(arc.angles(3)) == (180, 270, 0)

    arc.start_angle = -90
    arc.end_angle = -180
    assert tuple(arc.angles(2)) == (270, 180)
    assert tuple(arc.angles(4)) == (270, 0, 90, 180)


def test_vertices():
    angles = [0, 45, 90, 135, -45, -90, -135, 180]
    arc = ConstructionArc(center=(1, 1))
    vertices = list(arc.vertices(angles))
    for v, a in zip(vertices, angles):
        a = math.radians(a)
        assert v.isclose(Vec2((1 + math.cos(a), 1 + math.sin(a))))


def test_tangents():
    angles = [0, 45, 90, 135, -45, -90, -135, 180]
    sin45 = math.sin(math.pi / 4)
    result = [
        (0, 1),
        (-sin45, sin45),
        (-1, 0),
        (-sin45, -sin45),
        (sin45, sin45),
        (1, 0),
        (sin45, -sin45),
        (0, -1),
    ]
    arc = ConstructionArc(center=(1, 1))
    vertices = list(arc.tangents(angles))
    for v, r in zip(vertices, result):
        assert v.isclose(Vec2(r))


def test_angle_span():
    assert ConstructionArc(start_angle=30, end_angle=270).angle_span == 240
    # crossing 0-degree:
    assert (
        ConstructionArc(
            start_angle=30, end_angle=270, is_counter_clockwise=False
        ).angle_span
        == 120
    )
    # crossing 0-degree:
    assert ConstructionArc(start_angle=300, end_angle=60).angle_span == 120
    assert (
        ConstructionArc(
            start_angle=300, end_angle=60, is_counter_clockwise=False
        ).angle_span
        == 240
    )


def test_arc_segment_count():
    radius = 100
    max_sagitta = 2
    assert arc_segment_count(radius, math.tau, max_sagitta) == 16
    alpha = math.tau / 16
    l2 = math.sin(alpha / 2) * radius
    sagitta = radius - math.sqrt(radius**2 - l2**2)
    assert max_sagitta / 2 < sagitta < max_sagitta


class TestArcSegmentCountErrors:
    def test_radius_zero(self):
        assert arc_segment_count(radius=0, angle=math.tau, sagitta=1) == 1

    def test_sagitta_gt_radius(self):
        assert arc_segment_count(radius=1, angle=math.tau, sagitta=2) == 1


def test_arc_chord_length_domain_error():
    radius = 0.1
    assert arc_chord_length(radius, radius * 4) == 0.0


@pytest.mark.parametrize(
    "r, s, e, sagitta, count",
    [
        (1, 0, 180, 0.35, 3),
        (1, 0, 180, 0.10, 5),
        (0, 0, 360, 0.10, 0),  # radius 0 works but yields nothing
        (-1, 0, 180, 0.35, 3),  # negative radius same as positive radius
        (1, 270, 90, 0.10, 5),  # start angle > end angle
        (1, 90, -90, 0.10, 5),
        (1, 0, 0, 0.10, 0),  # angle span 0 works but yields nothing
        (1, -45, -45, 0.10, 0),
    ],
)
def test_flattening(r, s, e, sagitta, count):
    arc = ConstructionArc((0, 0), r, s, e)
    assert len(list(arc.flattening(sagitta))) == count


@pytest.mark.parametrize("p", [(2, 0), (2, 2), (0, 2), (2, -2), (0, -2)])
def test_point_is_in_arc_range(p):
    """
    Test if the angle defined by arc.center and point "p" is in the range
    arc.start_angle to arc.end_angle:
    """
    arc = ConstructionArc((0, 0), 1, -90, 90)
    assert arc._is_point_in_arc_range(Vec2(p)) is True


@pytest.mark.parametrize("p", [(-2, 0), (-2, 2), (-2, -2)])
def test_point_is_not_in_arc_range(p):
    """
    Test if the angle defined by arc.center and point "p" is NOT in the range
    arc.start_angle to arc.end_angle:
    """
    arc = ConstructionArc((0, 0), 1, -90, 90)
    assert arc._is_point_in_arc_range(Vec2(p)) is False


@pytest.mark.parametrize(
    "s, e",
    [
        [(0, 0), (2, 0)],  # touches the arc
        [(0, 0), (3, 0)],  # intersect
        [(0, 0), (0, 2)],  # touches the arc
        [(0, 0), (0, 3)],  # intersect
        [(0, 0), (2, 2)],  # intersect
        [(0, -1), (2, -1)],  # intersect
    ],
)
def test_arc_intersect_line_in_one_point(s, e):
    arc = ConstructionArc((0, 0), 2, -90, 90)
    assert len(arc.intersect_line(ConstructionLine(s, e))) == 1


@pytest.mark.parametrize(
    "s, e",
    [
        [(-2, 0), (2, 0)],  # touches
        [(-2, 1), (2, 1)],  # intersect
    ],
)
def test_arc_intersect_line_in_two_points(s, e):
    arc = ConstructionArc((0, 0), 2, 0, 180)
    assert len(arc.intersect_line(ConstructionLine(s, e))) == 2


@pytest.mark.parametrize(
    "s, e",
    [
        [(0, 2), (1, 2)],
        [(2, 0), (2, 1)],
        [(1, 1), (2, 2)],
    ],
)
def test_arc_does_not_intersect_line(s, e):
    arc = ConstructionArc((0, 0), 1, 0, 90)
    assert len(arc.intersect_line(ConstructionLine(s, e))) == 0


@pytest.mark.parametrize(
    "c, r",
    [
        [(0.0, 1.0), 1.0],
        [(0.0, 0.5), 0.5],
        [(2.0, 0.0), 1.0],
    ],
)
def test_arc_intersect_circle_in_one_point(c, r):
    arc = ConstructionArc((0, 0), 1, -90, 90)
    assert len(arc.intersect_circle(ConstructionCircle(c, r))) == 1


@pytest.mark.parametrize(
    "c, r",
    [
        [(1.0, 0.0), 1.0],
        [(0.5, 0.0), 1.0],
    ],
)
def test_arc_intersect_circle_in_two_points(c, r):
    arc = ConstructionArc((0, 0), 1, -90, 90)
    assert len(arc.intersect_circle(ConstructionCircle(c, r))) == 2


@pytest.mark.parametrize(
    "c, r",
    [
        [(0.0, 0.0), 0.5],  # concentric circle
        [(0.0, 0.0), 1.0],  # concentric circle
        [(0.0, 0.0), 2.0],  # concentric circle
        [(2.0, 0.0), 0.5],  # ) O
    ],
)
def test_arc_does_not_intersect_circle(c, r):
    arc = ConstructionArc((0, 0), 1, -90, 90)
    assert len(arc.intersect_circle(ConstructionCircle(c, r))) == 0


@pytest.mark.parametrize(
    "c, r, s, e",
    [
        [(2.0, 0.0), 1.0, 90, 270],  # touches in one point: )(
        [(1.5, 0.0), 1.0, 90, 180],  # intersect
    ],
)
def test_arc_intersect_arc_in_one_point(c, r, s, e):
    arc = ConstructionArc((0, 0), 1, -90, 90)
    assert len(arc.intersect_arc(ConstructionArc(c, r, s, e))) == 1


@pytest.mark.parametrize(
    "c, r, s, e",
    [
        [(0.5, 0.0), 1.0, 90, 270],  # intersect
        [(1.5, 0.0), 1.0, 90, 270],  # intersect
    ],
)
def test_arc_intersect_arc_in_two_points(c, r, s, e):
    arc = ConstructionArc((0, 0), 1, -90, 90)
    assert len(arc.intersect_arc(ConstructionArc(c, r, s, e))) == 2


@pytest.mark.parametrize(
    "c, r, s, e",
    [
        [(0.0, 0.0), 1.0, 90, 270],  # concentric arcs
        [(-0.5, 0.0), 1.0, 90, 270],  # insect circle but not arc: ( )
    ],
)
def test_arc_does_not_intersect_arc(c, r, s, e):
    arc = ConstructionArc((0, 0), 1, -90, 90)
    assert len(arc.intersect_arc(ConstructionArc(c, r, s, e))) == 0



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
import pytest
import ezdxf
from ezdxf import xref, units, const


@pytest.mark.parametrize("fmt", ["asc", "bin"])
@pytest.mark.parametrize("dxfversion", const.versions_supported_by_new)
def test_dxf_info_scanning(dxfversion, fmt, tmp_path):
    drawing_units = units.MM
    doc = ezdxf.new(dxfversion=dxfversion, units=drawing_units)
    doc.encoding = "cp1253"  # Greek
    doc.header["$INSBASE"] = (10, 20, 30)
    filename = tmp_path / f"{dxfversion}.dxf"
    doc.saveas(filename, fmt=fmt)

    info = xref.dxf_info(filename)
    assert info.version == dxfversion
    assert info.encoding == "cp1253"
    assert info.insert_base.isclose((10, 20, 30))

    # DXF R12 does not support the $INSUNITS header variable (checked by BricsCAD 2023)
    if dxfversion == const.DXF12:
        drawing_units = 0
    assert info.insert_units == drawing_units


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
import math

import ezdxf
from ezdxf.addons import r12export
from ezdxf import entities as ents
from ezdxf.render import forms
from ezdxf.tools.text import MTextEditor
from ezdxf.math import Vec2

DXFATTRIBS = {
    "layer": "EZDXF",
    "color": 144,
}


def test_export_empty_doc():
    doc = ezdxf.new("R2018")
    doc_r12 = r12export.convert(doc)
    assert doc_r12.dxfversion == ezdxf.DXF12


def test_export_dxf_primitives():
    doc = ezdxf.new("R2018")
    msp = doc.modelspace()
    msp.add_point((0, 0))
    msp.add_text("MyText")
    msp.add_line((0, 0), (1, 0))
    msp.add_circle((0, 0), radius=1)
    msp.add_arc((0, 0), radius=1, start_angle=30, end_angle=150)

    doc_r12 = r12export.convert(doc)
    assert len(doc_r12.modelspace()) == len(msp)


def test_export_lwpolyline_as_polyline():
    points = [(0, 0), (1, 0, 0.7), (2, 0), (2, 2)]
    doc = ezdxf.new("R2018")
    msp = doc.modelspace()
    msp.add_lwpolyline(points, format="xyb", dxfattribs=DXFATTRIBS)

    doc_r12 = r12export.convert(doc)
    polyline: ents.Polyline = doc_r12.modelspace()[0]
    assert isinstance(polyline, ents.Polyline)
    assert polyline.is_2d_polyline is True
    assert len(polyline.vertices) == 4
    assert polyline.vertices[1].dxf.bulge == 0.7
    assert polyline.dxf.layer == "EZDXF"
    assert polyline.dxf.color == 144


def test_export_mesh_as_polyface_mesh():
    doc = ezdxf.new("R2018")
    msp = doc.modelspace()
    cube = forms.cube()
    cube.render_mesh(msp, dxfattribs=DXFATTRIBS)

    doc_r12 = r12export.convert(doc)
    polyface: ents.Polyface = doc_r12.modelspace()[0]
    assert isinstance(polyface, ents.Polyface)
    assert polyface.is_poly_face_mesh is True
    assert polyface.dxf.layer == "EZDXF"
    assert polyface.dxf.color == 144


def test_export_ellipse_as_3d_polyline():
    doc = ezdxf.new("R2018")
    msp = doc.modelspace()
    msp.add_ellipse(
        (0, 0), (3, 0), 0.5, start_param=0, end_param=math.pi, dxfattribs=DXFATTRIBS
    )

    doc_r12 = r12export.convert(doc)
    polyline: ents.Polyline = doc_r12.modelspace()[0]
    assert isinstance(polyline, ents.Polyline)
    assert polyline.is_3d_polyline is True
    assert len(polyline.vertices) > 10
    assert polyline.dxf.layer == "EZDXF"
    assert polyline.dxf.color == 144


def test_export_spline_as_3d_polyline():
    doc = ezdxf.new("R2018")
    msp = doc.modelspace()
    msp.add_spline(fit_points=[(0, 0, 0), (2, 1, 1), (3, 5, -1)], dxfattribs=DXFATTRIBS)

    doc_r12 = r12export.convert(doc)
    polyline: ents.Polyline = doc_r12.modelspace()[0]
    assert isinstance(polyline, ents.Polyline)
    assert polyline.is_3d_polyline is True
    assert len(polyline.vertices) > 10
    assert polyline.dxf.layer == "EZDXF"
    assert polyline.dxf.color == 144



def test_export_proxy_graphic():
    from ezdxf.lldxf.tags import Tags
    from ezdxf.proxygraphic import load_proxy_graphic

    doc = ezdxf.new("R2018")
    msp = doc.modelspace()
    proxy_entity = ents.ACADProxyEntity.new()
    proxy_entity.proxy_graphic = load_proxy_graphic(Tags.from_text(DATA))
    msp.add_entity(proxy_entity)

    doc_r12 = r12export.convert(doc)
    assert len(doc_r12.modelspace()) == 9


def test_export_mtext():
    doc = ezdxf.new("R2018")
    msp = doc.modelspace()
    editor = MTextEditor()
    editor.append("LINE0\n")
    editor.font("Open Sans")
    editor.append("LINE1")
    msp.add_mtext(editor.text, dxfattribs=DXFATTRIBS)

    doc_r12 = r12export.convert(doc)
    text0, text1 = doc_r12.modelspace()

    assert text0.dxf.text == "LINE0"
    assert text1.dxf.text == "LINE1"
    assert doc_r12.styles.has_entry("MTXPL_OPEN_SANS")
    assert doc_r12.styles.has_entry("MTXPL_DEJAVU_SANS_CONDENSED")  # default font for TXT


def test_export_virtual_entities():
    doc = ezdxf.new(setup=True)
    mleaderstyle = doc.mleader_styles.duplicate_entry("Standard", "EZDXF")
    mleaderstyle.set_mtext_style("OpenSans")
    msp = doc.modelspace()
    ml_builder = msp.add_multileader_mtext("EZDXF")
    ml_builder.quick_leader(
        "Line1\nLine2",
        target=Vec2(40, 15),
        segment1=Vec2.from_deg_angle(45, 14),
    )
    doc_r12 = r12export.convert(doc)
    assert len(doc_r12.modelspace()) == 5


def test_export_hatch_pattern_fill():
    doc = ezdxf.new()
    msp = doc.modelspace()
    hatch = msp.add_hatch()
    hatch.paths.add_polyline_path([(0, 0), (10, 0), (10, 10), (0, 10)])
    hatch.set_pattern_fill("ANSI31", scale=0.5)

    doc_r12 = r12export.convert(doc)
    insert = doc_r12.modelspace()[0]
    block = doc_r12.blocks.get(insert.dxf.name)
    lines = block.query("LINE")
    assert len(lines) > 8


DATA = """160
968
310
C80300000D000000540000002000000002000000033E695D8B227240B00D3CF1FB7B5540000000000000000082C85BAC2FDE7240FB1040429FB05740000000000000000000000000000000000000000000000000000000000000F03F5400000020000000020000004AF9442AE7FA60405A2D686189715A4000000000000000
310
00C0DC003571AE5F40043422DDA4515D40000000000000000000000000000000000000000000000000000000000000F03F64000000040000001EA72DF9806A69402CE3B4E7B59D34400000000000000000770FBC9D50855E4000000000000000000000000000000000000000000000F03FB634003D352CE93FB1DDE561C5C1
310
E33F00000000000000000418DC3967E1F83F000000000C0000001200000000000000D0000000260000001F8BC5F8B8B46A40197732241FF06140000000000000000000000000000000000000000000000000000000000000F03F0943D77B25BDEF3F417457E0C451C0BF00000000000000003100370032002C003400320000
310
00000006000000010000000000000000000440000000000000F03F0000000000000000000000000000F03F00000000000000000000000000000000000000000000000000000000000000000000000041007200690061006C00000061007200690061006C002E007400740066000000000000000C00000012000000FF7F0000
310
6400000004000000813C33FBB3606A400278BF21B8F4614000000000000000009AEFA7C64B37034000000000000000000000000000000000000000000000F03F0943D77B25BDEF3F437457E0C451C0BF0000000000000000182D4454FB210940000000000C00000010000000010000000C0000001700000000000000540000
310
0020000000020000001EA72DF9806A69402CE3B4E7B59D344000000000000000001EA72DF9806A69402CE3B4E7B59D3440000000000000000000000000000000000000000000000000000000000000F03F540000002000000002000000B296839B8D1A724001000000F06355400000000000000000B296839B8D1A72400100
310
0000F0635540000000000000000000000000000000000000000000000000000000000000F03F540000002000000002000000632D073753076140FFFFFFFF2F525A400000000000000000632D073753076140FFFFFFFF2F525A40000000000000000000000000000000000000000000000000000000000000F03F5400000020
310
000000020000000960E446A3456F405AF2DBF448AB604000000000000000000960E446A3456F405AF2DBF448AB6040000000000000000000000000000000000000000000000000000000000000F03F
"""

if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
# Created 2019-02-13
import pytest
import copy
from ezdxf.lldxf.extendedtags import ExtendedTags
from ezdxf.entities.appdata import AppData


class TagWriter:
    """Mockup"""

    def __init__(self):
        self.tags = []

    def write_tags(self, tags):
        self.tags.append(tags)


@pytest.fixture
def tags():
    return ExtendedTags.from_text(APPDATA)


def test_app_data_get(tags):
    appdata = AppData()
    appdata.set(tags.appdata[0])
    mozman = appdata.get("MOZMAN")
    assert mozman[0] == (102, "{MOZMAN")
    assert mozman[-1] == (102, "}")


def test_app_data_add():
    appdata = AppData()
    appdata.add(
        "XXX",
        [
            (40, 3),
            (70, 19),
            (1, "Text"),
        ],
    )
    assert "XXX" in appdata
    data = appdata.get("XXX")
    assert len(data) == 5
    # is DXFTag()
    assert data[0].code == 102
    assert data[0].value == "{XXX"
    assert data[-1].code == 102
    assert data[-1].value == "}"


def test_app_data_add_data():
    appdata = AppData()
    appdata.add(
        "XXX",
        [
            (40, 3),
            (70, 19),
            (1, "Text"),
        ],
    )
    assert "XXX" in appdata
    data = appdata.get("XXX")
    assert len(data) == 5
    # is DXFTag()
    assert data[0].code == 102
    assert data[0].value == "{XXX"
    assert data[-1].code == 102
    assert data[-1].value == "}"


def test_app_data_delete(tags):
    appdata = AppData()
    appdata.set(tags.appdata[0])
    assert "MOZMAN" in appdata
    appdata.discard("MOZMAN")
    assert "MOZMAN" not in appdata
    # nothing happens if appid does not exist
    appdata.discard("MOZMAN")


def test_app_data_dxf_export(tags):
    appdata = AppData()
    appdata.set(tags.appdata[0])
    appdata.add(
        "XXX",
        [
            (40, 3),
            (70, 19),
            (1, "Text"),
        ],
    )
    assert len(appdata) == 2
    tagwriter = TagWriter()
    appdata.export_dxf(tagwriter)

    assert len(tagwriter.tags) == 2
    tags1, tags2 = tagwriter.tags
    assert tags1[0] == (102, "{MOZMAN")
    assert tags1[-1] == (102, "}")
    assert tags2[0] == (102, "{XXX")
    assert tags2[-1] == (102, "}")


def test_clone(tags):
    appdata = AppData()
    appdata.set(tags.appdata[0])
    new_appdata = copy.deepcopy(appdata)
    new_appdata.add(
        "MOZMAN",
        [
            (1, "Text"),
        ],
    )
    d1 = appdata.get("MOZMAN")
    d2 = new_appdata.get("MOZMAN")
    assert len(d1) == 4
    assert len(d2) == 3
    assert d1[1] == (40, 1.0)
    assert d2[1] == (1, "Text")


APPDATA = """0
DXFENTITY
5
FFFF
102
{MOZMAN
40
1.0
70
9
102
}
"""

if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import (
    Optional,
    Iterable,
    Iterator,
    Sequence,
    NamedTuple,
    Callable,
    TYPE_CHECKING,
)
import abc

from ezdxf.math import (
    Matrix44,
    Vec2,
    BoundingBox2d,
    UVec,
    is_convex_polygon_2d,
    is_axes_aligned_rectangle_2d,
)
from ezdxf.npshapes import NumpyPath2d, NumpyPoints2d

if TYPE_CHECKING:
    from ezdxf.math.clipping import Clipping

__all__ = [
    "ClippingShape",
    "ClippingPortal",
    "ClippingRect",
    "ConvexClippingPolygon",
    "InvertedClippingPolygon",
    "MultiClip",
    "find_best_clipping_shape",
    "make_inverted_clipping_shape",
]


class ClippingShape(abc.ABC):
    """The ClippingShape defines a single clipping path and executes the clipping on
    basic geometries:

    - point: a single point
    - line: a line between two vertices
    - polyline: open polyline with one or more straight line segments
    - polygon: closed shape with straight line as edges
    - path: open shape with straight lines and Bezier-curves as segments
    - filled-path: closed shape with straight lines and Bezier-curves as edges

    Difference between open and closed shapes:

        - an open shape is treated as a linear shape without a filling
        - clipping an open shape returns one or more open shapes
        - a closed shape is treated as a filled shape, where the first vertex is
          coincident to the last vertex.
        - clipping a closed shape returns one or more closed shapes

    Notes:

        An arbitrary clipping polygon can split any basic geometry (except point) into
        multiple parts.

        All current implemented clipping algorithms flatten Bezier-curves into polylines.

    """

    @abc.abstractmethod
    def bbox(self) -> BoundingBox2d: ...

    @abc.abstractmethod
    def is_completely_inside(self, other: BoundingBox2d) -> bool: ...

    # returning False means: I don't know!

    @abc.abstractmethod
    def is_completely_outside(self, other: BoundingBox2d) -> bool: ...

    @abc.abstractmethod
    def clip_point(self, point: Vec2) -> Optional[Vec2]: ...

    @abc.abstractmethod
    def clip_line(self, start: Vec2, end: Vec2) -> Sequence[tuple[Vec2, Vec2]]: ...

    @abc.abstractmethod
    def clip_polyline(self, points: NumpyPoints2d) -> Sequence[NumpyPoints2d]: ...

    @abc.abstractmethod
    def clip_polygon(self, points: NumpyPoints2d) -> Sequence[NumpyPoints2d]: ...

    @abc.abstractmethod
    def clip_paths(
        self, paths: Iterable[NumpyPath2d], max_sagitta: float
    ) -> Iterator[NumpyPath2d]: ...

    @abc.abstractmethod
    def clip_filled_paths(
        self, paths: Iterable[NumpyPath2d], max_sagitta: float
    ) -> Iterator[NumpyPath2d]: ...


class ClippingStage(NamedTuple):
    portal: ClippingShape
    transform: Matrix44 | None


class ClippingPortal:
    """The ClippingPortal manages a clipping path stack."""

    def __init__(self) -> None:
        self._stages: list[ClippingStage] = []

    @property
    def is_active(self) -> bool:
        return bool(self._stages)

    def push(self, portal: ClippingShape, transform: Matrix44 | None) -> None:
        self._stages.append(ClippingStage(portal, transform))

    def pop(self) -> None:
        if self._stages:
            self._stages.pop()

    def foreach_stage(self, command: Callable[[ClippingStage], bool]) -> None:
        for stage in self._stages[::-1]:
            if not command(stage):
                return

    def clip_point(self, point: Vec2) -> Optional[Vec2]:
        result: Vec2 | None = point

        def do(stage: ClippingStage) -> bool:
            nonlocal result
            assert result is not None
            if stage.transform:
                result = Vec2(stage.transform.transform(result))
            result = stage.portal.clip_point(result)
            return result is not None

        self.foreach_stage(do)
        return result

    def clip_line(self, start: Vec2, end: Vec2) -> list[tuple[Vec2, Vec2]]:
        def do(stage: ClippingStage) -> bool:
            lines = list(result)
            result.clear()
            for s, e in lines:
                if stage.transform:
                    s, e = stage.transform.fast_2d_transform((s, e))
                result.extend(stage.portal.clip_line(s, e))
            return bool(result)

        result = [(start, end)]
        self.foreach_stage(do)
        return result

    def clip_polyline(self, points: NumpyPoints2d) -> list[NumpyPoints2d]:
        def do(stage: ClippingStage) -> bool:
            polylines = list(result)
            result.clear()
            for polyline in polylines:
                if stage.transform:
                    polyline.transform_inplace(stage.transform)
                result.extend(stage.portal.clip_polyline(polyline))
            return bool(result)

        result = [points]
        self.foreach_stage(do)
        return result

    def clip_polygon(self, points: NumpyPoints2d) -> list[NumpyPoints2d]:
        def do(stage: ClippingStage) -> bool:
            polygons = list(result)
            result.clear()
            for polygon in polygons:
                if stage.transform:
                    polygon.transform_inplace(stage.transform)
                result.extend(stage.portal.clip_polygon(polygon))
            return bool(result)

        result = [points]
        self.foreach_stage(do)
        return result

    def clip_paths(
        self, paths: Iterable[NumpyPath2d], max_sagitta: float
    ) -> list[NumpyPath2d]:
        def do(stage: ClippingStage) -> bool:
            paths = list(result)
            result.clear()
            for path in paths:
                if stage.transform:
                    path.transform_inplace(stage.transform)
            result.extend(stage.portal.clip_paths(paths, max_sagitta))
            return bool(result)

        result = list(paths)
        self.foreach_stage(do)
        return result

    def clip_filled_paths(
        self, paths: Iterable[NumpyPath2d], max_sagitta: float
    ) -> list[NumpyPath2d]:
        def do(stage: ClippingStage) -> bool:
            paths = list(result)
            result.clear()
            for path in paths:
                if stage.transform:
                    path.transform_inplace(stage.transform)
            result.extend(stage.portal.clip_filled_paths(paths, max_sagitta))
            return bool(result)

        result = list(paths)
        self.foreach_stage(do)
        return result

    def transform_matrix(self, m: Matrix44) -> Matrix44:
        for _, transform in self._stages[::-1]:
            if transform is not None:
                m @= transform
        return m


class ClippingPolygon(ClippingShape):
    """Represents an arbitrary polygon as clipping shape.  Removes the geometry
    outside the clipping polygon.

    """

    def __init__(self, bbox: BoundingBox2d, clipper: Clipping) -> None:
        if not bbox.has_data:
            raise ValueError("clipping box not detectable")
        self._bbox = bbox
        self.clipper = clipper

    def bbox(self) -> BoundingBox2d:
        return self._bbox

    def clip_point(self, point: Vec2) -> Optional[Vec2]:
        is_inside = self.clipper.is_inside(Vec2(point))
        if not is_inside:
            return None
        return point

    def clip_line(self, start: Vec2, end: Vec2) -> Sequence[tuple[Vec2, Vec2]]:
        return self.clipper.clip_line(start, end)

    def clip_polyline(self, points: NumpyPoints2d) -> Sequence[NumpyPoints2d]:
        clipper = self.clipper
        if len(points) == 0:
            return tuple()
        polyline_bbox = BoundingBox2d(points.extents())
        if self.is_completely_outside(polyline_bbox):
            return tuple()
        if self.is_completely_inside(polyline_bbox):
            return (points,)
        return [
            NumpyPoints2d(part)
            for part in clipper.clip_polyline(points.vertices())
            if len(part) > 0
        ]

    def clip_polygon(self, points: NumpyPoints2d) -> Sequence[NumpyPoints2d]:
        clipper = self.clipper
        if len(points) < 2:
            return tuple()
        polygon_bbox = BoundingBox2d(points.extents())
        if self.is_completely_outside(polygon_bbox):
            return tuple()
        if self.is_completely_inside(polygon_bbox):
            return (points,)
        return [
            NumpyPoints2d(part)
            for part in clipper.clip_polygon(points.vertices())
            if len(part) > 0
        ]

    def clip_paths(
        self, paths: Iterable[NumpyPath2d], max_sagitta: float
    ) -> Iterator[NumpyPath2d]:
        clipper = self.clipper
        for path in paths:
            for sub_path in path.sub_paths():
                path_bbox = BoundingBox2d(sub_path.control_vertices())
                if not path_bbox.has_data:
                    continue
                if self.is_completely_inside(path_bbox):
                    yield sub_path
                    continue
                if self.is_completely_outside(path_bbox):
                    continue
                polyline = Vec2.list(sub_path.flattening(max_sagitta, segments=4))
                for part in clipper.clip_polyline(polyline):
                    if len(part) > 0:
                        yield NumpyPath2d.from_vertices(part, close=False)

    def clip_filled_paths(
        self, paths: Iterable[NumpyPath2d], max_sagitta: float
    ) -> Iterator[NumpyPath2d]:
        clipper = self.clipper
        for path in paths:
            for sub_path in path.sub_paths():
                if len(sub_path) < 2:
                    continue
                path_bbox = BoundingBox2d(sub_path.control_vertices())
                if self.is_completely_inside(path_bbox):
                    yield sub_path
                    continue
                if self.is_completely_outside(path_bbox):
                    continue
                for part in clipper.clip_polygon(
                    Vec2.list(sub_path.flattening(max_sagitta, segments=4))
                ):
                    if len(part) > 0:
                        yield NumpyPath2d.from_vertices(part, close=True)


class ClippingRect(ClippingPolygon):
    """Represents a rectangle as clipping shape where the edges are parallel to
    the x- and  y-axis of the coordinate system.  Removes the geometry outside the
    clipping rectangle.

    """

    def __init__(self, vertices: Iterable[UVec]) -> None:
        from ezdxf.math.clipping import ClippingRect2d

        polygon = Vec2.list(vertices)
        bbox = BoundingBox2d(polygon)
        if not bbox.has_data:
            raise ValueError("clipping box not detectable")
        size: Vec2 = bbox.size
        self.remove_all = size.x * size.y < 1e-9

        super().__init__(bbox, ClippingRect2d(bbox.extmin, bbox.extmax))

    def is_completely_inside(self, other: BoundingBox2d) -> bool:
        return self._bbox.contains(other)

    def is_completely_outside(self, other: BoundingBox2d) -> bool:
        return not self._bbox.has_intersection(other)

    def clip_point(self, point: Vec2) -> Optional[Vec2]:
        if self.remove_all:
            return None
        return super().clip_point(point)

    def clip_line(self, start: Vec2, end: Vec2) -> Sequence[tuple[Vec2, Vec2]]:
        if self.remove_all:
            return tuple()
        return self.clipper.clip_line(start, end)

    def clip_polyline(self, points: NumpyPoints2d) -> Sequence[NumpyPoints2d]:
        if self.remove_all:
            return (NumpyPoints2d(tuple()),)
        return super().clip_polyline(points)

    def clip_polygon(self, points: NumpyPoints2d) -> Sequence[NumpyPoints2d]:
        if self.remove_all:
            return (NumpyPoints2d(tuple()),)
        return super().clip_polygon(points)

    def clip_paths(
        self, paths: Iterable[NumpyPath2d], max_sagitta: float
    ) -> Iterator[NumpyPath2d]:
        if self.remove_all:
            return iter(tuple())
        return super().clip_paths(paths, max_sagitta)

    def clip_filled_paths(
        self, paths: Iterable[NumpyPath2d], max_sagitta: float
    ) -> Iterator[NumpyPath2d]:
        if self.remove_all:
            return iter(tuple())
        return super().clip_filled_paths(paths, max_sagitta)


class ConvexClippingPolygon(ClippingPolygon):
    """Represents an arbitrary convex polygon as clipping shape.  Removes the geometry
    outside the clipping polygon.

    """

    def __init__(self, vertices: Iterable[UVec]) -> None:
        from ezdxf.math.clipping import ConvexClippingPolygon2d

        polygon = Vec2.list(vertices)
        super().__init__(BoundingBox2d(polygon), ConvexClippingPolygon2d(polygon))

    def is_completely_inside(self, other: BoundingBox2d) -> bool:
        return False  # I don't know!

    def is_completely_outside(self, other: BoundingBox2d) -> bool:
        return not self._bbox.has_intersection(other)


class ConcaveClippingPolygon(ClippingPolygon):
    """Represents an arbitrary concave polygon as clipping shape.  Removes the geometry
    outside the clipping polygon.

    """

    def __init__(self, vertices: Iterable[UVec]) -> None:
        from ezdxf.math.clipping import ConcaveClippingPolygon2d

        polygon = Vec2.list(vertices)
        super().__init__(BoundingBox2d(polygon), ConcaveClippingPolygon2d(polygon))

    def is_completely_inside(self, other: BoundingBox2d) -> bool:
        return False  # I don't know!

    def is_completely_outside(self, other: BoundingBox2d) -> bool:
        return not self._bbox.has_intersection(other)


class InvertedClippingPolygon(ClippingPolygon):
    """Represents an arbitrary inverted clipping polygon.  Removes the geometry
    inside the clipping polygon.

    .. Important:: 
    
        The `outer_bounds` must be larger than the content to clip to work correctly.

    """

    def __init__(self, vertices: Iterable[UVec], outer_bounds: BoundingBox2d) -> None:
        from ezdxf.math.clipping import InvertedClippingPolygon2d

        polygon = Vec2.list(vertices)
        super().__init__(outer_bounds, InvertedClippingPolygon2d(polygon, outer_bounds))

    def is_completely_inside(self, other: BoundingBox2d) -> bool:
        # returning False means: I don't know!
        return False  # not easy to detect

    def is_completely_outside(self, other: BoundingBox2d) -> bool:
        return not self._bbox.has_intersection(other)


def find_best_clipping_shape(polygon: Iterable[UVec]) -> ClippingShape:
    """Returns the best clipping shape for the given clipping polygon.

    The function analyses the given polygon (rectangular, convex or concave polygon, ...)
    and returns the optimized (fastest) clipping shape.

    Args:
        polygon: clipping polygon as iterable vertices

    """
    points = Vec2.list(polygon)
    if is_axes_aligned_rectangle_2d(points):
        return ClippingRect(points)
    elif is_convex_polygon_2d(points, strict=False):
        return ConvexClippingPolygon(points)
    return ConcaveClippingPolygon(points)


def make_inverted_clipping_shape(
    polygon: Iterable[UVec], outer_bounds: BoundingBox2d
) -> ClippingShape:
    """Returns an inverted clipping shape that removes the geometry inside the clipping
    polygon and beyond the outer bounds.

    Args:
        polygon: clipping polygon as iterable vertices
        outer_bounds: outer bounds of the clipping shape

    """

    return InvertedClippingPolygon(polygon, outer_bounds)



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
import pytest

import ezdxf
from ezdxf.acis import api as acis


def test_create_a_new_cache():
    cache = acis.AcisCache()

    assert len(cache) == 0


def test_get_bodies_from_sat_data(cube_sat: str):
    cache = acis.AcisCache()
    bodies = cache.get_bodies(cube_sat)

    assert len(bodies) == 1
    assert isinstance(bodies[0], acis.Body)
    assert len(cache) == 1

def test_empty_data_does_not_create_cache_entries():
    cache = acis.AcisCache()
    bodies = cache.get_bodies([])

    assert len(bodies) == 0
    assert len(cache) == 0


def test_get_bodies_from_sab_data(cube_sab: bytes):
    cache = acis.AcisCache()
    bodies = cache.get_bodies(cube_sab)

    assert len(bodies) == 1
    assert isinstance(bodies[0], acis.Body)
    assert len(cache) == 1


def test_add_only_unique_entries(cube_sat: str):
    cache = acis.AcisCache()
    cache.get_bodies(cube_sat)
    cache.get_bodies(cube_sat)

    assert len(cache) == 1, "should not create duplicate entries"


hash_data = acis.AcisCache.hash_data


class TestHashData:
    def test_empty_string(self):
        if ezdxf.PYPY:
            assert hash_data("") != 0
        else:
            assert hash_data("") == 0

    def test_string(self):
        h = hash_data("abc")
        assert isinstance(h, int)
        assert h != 0
        assert h == hash_data("abc")

    def test_strings(self):
        data = ["abc", "def"]
        h = hash_data(data)
        assert isinstance(h, int)
        assert h != 0
        assert h == hash_data(("abc", "def"))

    def test_empty_string_list(self):
        data = []
        h = hash_data(data)
        assert isinstance(h, int)
        assert h != 0
        assert h == hash_data(data)

        # surprise?
        assert hash(tuple()) == hash(tuple())
        assert id(tuple()) == id(tuple())

    def test_empty_bytes(self):
        if ezdxf.PYPY:
            assert hash_data(b"") != 0
        else:
            assert hash_data(b"") == 0

    def test_bytes(self):
        data = b"\x07\x08\x09"
        h = hash_data(data)
        assert isinstance(h, int)
        assert h != 0
        assert h == hash_data(b"\x07\x08\x09")

    def test_bytearray(self):
        data = bytearray([7, 8, 9])
        h = hash_data(data)
        assert isinstance(h, int)
        assert h != 0
        assert h == hash_data(b"\x07\x08\x09")


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020-2022, Manfred Moitzi
# License: MIT License
from typing import Iterable, Iterator
import ezdxf
from ezdxf.math import UVec
from ._linetypes import _LineTypeRenderer, LineSegment

if ezdxf.options.use_c_ext:
    try:
        from ezdxf.acc.linetypes import _LineTypeRenderer  # type: ignore
    except ImportError:
        pass


class LineTypeRenderer(_LineTypeRenderer):
    def line_segments(self, vertices: Iterable[UVec]) -> Iterator[LineSegment]:
        last = None
        for vertex in vertices:
            if last is not None:
                yield from self.line_segment(last, vertex)
            last = vertex



#  Copyright (c) 2021-2023, Manfred Moitzi
#  License: MIT License
from typing import cast
import pytest
import ezdxf
from ezdxf import zoom, bbox


class TestModelSpace:
    @pytest.fixture(scope="class")
    def msp(self):
        doc = ezdxf.new()
        msp_ = doc.modelspace()
        msp_.add_point((-25, -25))
        msp_.add_point((25, 25))
        msp_.add_point((-100, -50))
        msp_.add_point((100, 50))
        return msp_

    def test_zoom_center(self, msp):
        zoom.center(msp, (10, 10), (10, 10))
        vp = msp.doc.viewports.get("*Active")[0]
        assert vp.dxf.center == (10, 10)
        assert vp.dxf.height == 10

    def test_zoom_extents(self, msp):
        zoom.extents(msp)
        vp = msp.doc.viewports.get("*Active")[0]
        assert vp.dxf.center == (0, 0)
        # 2:1 = 200 / 2 = 100 == height
        assert vp.dxf.height == 100

    def test_zoom_extents_factor_2(self, msp):
        zoom.extents(msp, factor=2)
        vp = msp.doc.viewports.get("*Active")[0]
        assert vp.dxf.center == (0, 0)
        # 2:1 = 200 / 2 = 100 * 2 == height
        assert vp.dxf.height == 200

    def test_zoom_window(self, msp):
        zoom.window(msp, (-30, -10), (0, 20))
        vp = msp.doc.viewports.get("*Active")[0]
        assert vp.dxf.center == (-15, 5)
        assert vp.dxf.height == 30

    def test_zoom_objects(self, msp):
        points = msp[:2]
        zoom.objects(msp, points)
        vp = msp.doc.viewports.get("*Active")[0]
        assert vp.dxf.center == (0, 0)
        assert vp.dxf.height == 50

    def test_zoom_objects_of_empty_set(self, msp):
        vp = msp.doc.viewports.get("*Active")[0]
        old_center = vp.dxf.center
        old_height = vp.dxf.height
        zoom.objects(msp, [])
        vp = msp.doc.viewports.get("*Active")[0]
        assert vp.dxf.center == old_center
        assert vp.dxf.height == old_height


class TestPaperSpace:
    @pytest.fixture(scope="class")
    def psp(self):
        doc = ezdxf.new()
        psp_ = cast("Paperspace", doc.layout("Layout1"))
        psp_.add_viewport(
            center=(40, 40),
            size=(40, 40),
            view_center_point=(0, 0),
            view_height=1,
        )
        psp_.add_point((5, 5))
        psp_.add_point((405, 305))
        return psp_

    def test_bbox_of_main_viewport(self, psp):
        psp.reset_main_viewport()
        vp = psp.main_viewport()
        box = bbox.extents([vp])
        assert box.has_data is True
        assert box.center.isclose(vp.dxf.center)
        assert box.size.x == pytest.approx(vp.dxf.width)
        assert box.size.y == pytest.approx(vp.dxf.height)

    def test_zoom_center(self, psp):
        zoom.center(psp, (10, 10), size=(20, 15))
        vp = psp.main_viewport()
        assert vp.dxf.center == (10, 10)
        assert vp.dxf.width == 20
        assert vp.dxf.height == 15

    def test_zoom_extents(self, psp):
        zoom.extents(psp)
        vp = psp.main_viewport()
        assert vp.dxf.center == (205, 155)
        assert vp.dxf.width == 400
        assert vp.dxf.height == 300

    def test_zoom_extents_factor_2(self, psp):
        zoom.extents(psp, factor=2)
        vp = psp.main_viewport()
        assert vp.dxf.center == (205, 155)
        assert vp.dxf.width == 800
        assert vp.dxf.height == 600

    def test_zoom_window(self, psp):
        zoom.window(psp, (-30, -10), (0, 20))
        vp = psp.main_viewport()
        assert vp.dxf.center == (-15, 5)
        assert vp.dxf.height == 30

    def test_zoom_objects_of_empty_set(self, psp):
        vp = psp.main_viewport()
        old_center = vp.dxf.center
        old_height = vp.dxf.height
        zoom.objects(psp, [])
        vp = psp.main_viewport()
        assert vp.dxf.center == old_center
        assert vp.dxf.height == old_height


if __name__ == "__main__":
    pytest.main([__file__])
